"""Geodetic transformations. This module holds many functions to peform common
transformations used in astrodynamics.

Copyright (C) {2017}  {Shankar Kulumani}

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
import numpy as np
from kinematics import attitude
from . import time, constants
import pdb

def lla2ecef(lat, lon, alt, r=6378.137, ee=8.1819190842622e-2):
    """Convert latitude, longitude, and altitude to the Earth centered
    Earth fixed frame (ECEF)

    Will convert between geodetic latitude, longitude, and altitude to the
    associated planet fixed reference frame.  This assumes an oblate spheroid
    model for the planetary body. It is applicable to any spherical body given
    the semi-major axis and the body eccentricity constant.

    Parameters:
    -----------
    lat : float
        Geodetic latitude (radians)
    lon : float
        Geodetic longitude (radians)
    alt : float
        height about the mean surface ellipsoid (WGS84 for Earth) (kilometers)

    Returns
    -------
    ecef : float ndarray (3,)
        Body fixed position vector in (kilometers)

    Notes
    -----
    We're assuming an oblate spheroid model for the Earth. To transform this
    vector to the inertial ECI frame all you need to do is rotate it by the
    Greenwich Mean Sideral time (GST) or use a fancy ECEF-ECI rotation which
    accounts for the Earth's precession/nutation.

    References
    ----------
    .. [1] BATE, Roger R, MUELLER, Donald D WHITE, Jerry E. Fundamentals of
    Astrodynamics. Courier Dover Publications, 1971.

    .. [2] Nice website to verify computations:
    http://www.oc.nps.edu/oc2902w/coord/llhxyz.htm

    Examples
    --------
    Some examples demonstrating the usage of this function

    >>> import numpy

    Convert the lattitude, longitude, and altitude of Washington, DC to it's
    equivalent ECEF representation.

    >>> lat, lon, alt = (38.8895 * np.pi / 180, -77.035 * np.pi / 180, 0.054)

    We need to make sure the inputs are in the correct units.

    >>> lla2ecef(lat, lon, alt)
    [1115.308, -4844.546, 3982.965]

    """
    # Normal distance from teh surface to the Z axis along the ellipsoid normal
    N = r / np.sqrt(1 - ee**2 * np.sin(lat)**2)

    x = (N + alt) * np.cos(lat) * np.cos(lon)
    y = (N + alt) * np.cos(lat) * np.sin(lon)
    z = ((1 - ee**2) * N + alt) * np.sin(lat)

    return np.array([x, y, z])


def eci2ecef(jd):
    """Rotation matrix to convert from ECI to ECEF

    Will gradually improve this function to include all of the Earth
    motion terms. For now, just use sidereal time to get the rotation matrix
    """

    gst, _ = time.gstlst(jd, 0)

    dcm = attitude.rot3(gst, 'r')

    return dcm


def site2ecef(lat, alt, lst, r=6378.137, ee=8.1819190842622e-2):
    """Calculate the site vector in the ECEF coordinate system.

    Author:   C2C Shankar Kulumani   USAFA/CS-19   719-333-4741

    Inputs:
        sitlat - site latitude (radians)
        lst - local sidereal time (radians)
        sitalt - site altitude (meters)

    Outputs:
        R_site - site vector in ECEF frame

    Globals:
        RE, EEsqrd

    Constants: None

    Coupling: None

    References:
        Astro 321 COMFIX
        Vallado Alg 47
    """
    N = r / np.sqrt(1 - ee**2 * np.sin(lat)**2)
    x = (N + alt) * np.cos(lat)
    z = ((1 - ee**2) * N + alt) * np.sin(lat)
    ecef = np.array([x * np.cos(lst), x * np.sin(lst), z])

    return ecef


def ecef2lla(ecef, r=6378.137, ee=8.1819190842622e-2):
    """Converts a ECEF vector to the equivalent Lat, longitude and Altitude
    above the reference ellipsoid

    """
    twopi = 2 * np.pi
    tol = 1e-6

    norm_vec = np.linalg.norm(ecef)

    temp = np.sqrt(ecef[0]**2 + ecef[1]**2)
    if temp < tol:
        rtasc = np.sign(ecef[2]) * np.pi * 0.5
    else:
        rtasc = np.arctan2(ecef[1], ecef[0])

    lon = rtasc
    lon = attitude.normalize(lon, 0, 2 * np.pi)

    decl = np.arcsin(ecef[2] / norm_vec)
    latgd = decl

    # iterate to find geodetic latitude
    i = 1
    olddelta = latgd + 10.0
    while np.absolute(olddelta - latgd) >= tol and i < 10:
        oldelta = latgd
        sintemp = np.sin(latgd)
        c = r / np.sqrt(1.0 - ee**2 * sintemp**2)
        latgd = np.arctan2(ecef[2] + c * ee**2 * sintemp, temp)
        i = i + 1

    # calculate the height
    if (np.pi / 2 - np.absolute(latgd)) > np.pi / 180:
        hellp = temp / np.cos(latgd) - c
    else:
        s = c * (1 - ee**2)
        hellp = ecef[2] / np.sin(latgd) - s

    latgc = gd2gc(latgd, ee**2)
    return latgc, latgd, lon, hellp


def gd2gc(latgd, eesqrd=0.081819221456**2):
    """Only valid for locations on the Earth's surface

    Vallado Example 3-1
    """
    latgc = np.arctan((1 - eesqrd) * np.tan(latgd))
    return latgc


def gc2gd(latgc, eesqrd=0.081819221456**2):
    """Only valid for locations on the Earth's surface

    Vallado Example 3-1
    """
    latgd = np.arctan(np.tan(latgc) / (1 - eesqrd))
    return latgd


def rv2rhoazel(r_sat_eci, v_sat_eci, lat, lon, alt, jd):
    """
    This function calculates the topcentric range,azimuth and elevation from
    the site vector and satellite position vector.

    Author:   C2C Shankar Kulumani   USAFA/CS-19   719-333-4741

    Inputs:
        sat_eci - satellite ECI position vector (km)
        site_eci - site ECI position vector (km)
        site_lat - site geodetic latitude (radians)
        site_lst - site local sidereal time (radians)

    Outputs:
        rho - range (km)
        az - asimuth (radians)
        el - elevation (radians)

    Globals: None

    Constants: None

    Coupling:

    References:
        Astro 321 Predict LSN 22
    """
    small = constants.small
    halfpi = constants.halfpi

    # get site in ECEF
    r_site_ecef = lla2ecef(lat, lon, alt)

    # convert sat and site to ecef
    dcm_eci2ecef = eci2ecef(jd)
    omega_earth = np.array([0, 0, constants.earth.omega])

    r_sat_ecef = dcm_eci2ecef.dot(r_sat_eci)
    v_sat_ecef = dcm_eci2ecef.dot(
        v_sat_eci) - np.cross(omega_earth, r_sat_ecef)

    # find relative vector in ecef frame
    rho_ecef = r_sat_ecef - r_site_ecef
    drho_ecef = v_sat_ecef - np.zeros_like(v_sat_ecef)  # site isn't moving
    rho = np.linalg.norm(rho_ecef)

    # convert to SEZ coordinate frame
    dcm_ecef2sez = attitude.rot2(np.pi / 2 - lat, 'r').dot(attitude.rot3(lon, 'r'))
    rho_sez = dcm_ecef2sez.dot(rho_ecef)
    drho_sez = dcm_ecef2sez.dot(drho_ecef)

    # find azimuth and eleveation
    temp = np.sqrt(rho_sez[0]**2 + rho_sez[1]**2)

    if temp < small:  # directly over the north pole
        el = np.sign(rho_sez[2]) * halfpi  # \pm 90 deg
    else:
        mag_rho_sez = np.linalg.norm(rho_sez)
        el = np.arcsin(rho_sez[2] / mag_rho_sez)

    if temp < small:
        az = np.arctan2(drho_sez[1], - drho_sez[0])
    else:
        az = np.arctan2(rho_sez[1] / temp, -rho_sez[0] / temp)

    # range, azimuth and elevation rates
    drho = np.dot(rho_sez, drho_sez) / rho

    if np.absolute(temp**2) > small:
        daz = (drho_sez[0] * rho_sez[1] - drho_sez[1] * rho_sez[0]) / temp**2
    else:
        daz = 0

    if np.absolute(temp) > small:
        dele = (drho_sez[2] - drho * np.sin(el)) / temp
    else:
        dele = 0
    
    az = attitude.normalize(az, 0, constants.twopi)
    return rho, az, el, drho, daz, dele


def rhoazel(sat_eci, site_eci, site_lat, site_lst):
    """
    This function calculates the topcentric range,azimuth and elevation from
    the site vector and satellite position vector.

    Author:   C2C Shankar Kulumani   USAFA/CS-19   719-333-4741

    Inputs:
        sat_eci - satellite ECI position vector (km)
        site_eci - site ECI position vector (km)
        site_lat - site geodetic latitude (radians)
        site_lst - site local sidereal time (radians)

    Outputs:
        rho - range (km)
        az - asimuth (radians)
        el - elevation (radians)

    Globals: None

    Constants: None

    Coupling: 

    References:
        Astro 321 Predict LSN 22 
    """

    site2sat_eci = sat_eci - site_eci

    site2sat_sez = attitude.rot3(-site_lst).dot(site2sat_eci)
    site2sat_sez = attitude.rot2(-(np.pi / 2 - site_lat)).dot(site2sat_sez)

    rho = np.linalg.norm(site2sat_sez)
    el = np.arcsin(site2sat_sez[2] / rho)
    az = attitude.normalize(np.arctan2(site2sat_sez[1], -site2sat_sez[0]), 0,
                            2 * np.pi)[0]

    return rho, az, el



"""
Do lots of functions for the planets, sun and/or moon
"""
from __future__ import absolute_import, division, print_function, unicode_literals
from collections import namedtuple
import numpy as np
from kinematics import attitude
from . import kepler, constants
import pdb

twopi = 2 * np.pi
deg2rad = np.pi / 180
rad2deg = 180 / np.pi
au2km = 149597870.0

COE = namedtuple('COE', ['p', 'ecc', 'inc', 'raan', 'argp', 'nu'])


def sun_earth_eci(jd):
    """This function calculates the Geocentric Equatorial position vector for
    the Sun given the Julian Date.  This is the low precision formula and is
    valid for years from 1950 to 2050.  Accuaracy of apparent coordinates is
    0.01 degrees.  Notice many of the calculations are performed in degrees,
    and are not changed until later.  This is due to the fact that the Almanac
    uses degrees exclusively in their formulations.

    Algorithm     : Calculate the several values needed to find the vector
                    Be careful of quadrant checks

    Author        : Capt Dave Vallado  USAFA/DFAS  719-472-4109  25 Aug 1988
    In Ada        : Dr Ron Lisowski    USAFA/DFAS  719-472-4110  17 May 1995
    In MatLab     : Dr Ron Lisowski    USAFA/DFAS  719-333-4109  10 Oct 2001
    In Python     : Shankar Kulumani   GWU         630-336-6257  19 Jun 2017

    Inputs        :
    JD          - Julian Date                            days from 4713 B.C.

    Outputs       :
    RSun        - IJK Position vector of the Sun         km
    RtAsc       - Right Ascension                        rad
    Decl        - Declination                            rad

    Locals        :
    MeanLong    - Mean Longitude
    MeanAnomaly - Mean anomaly
    N           - Number of days from 1 Jan 2000
    EclpLong    - Ecliptic longitude
    Obliquity   - Mean Obliquity of the Ecliptic

    Constants     :
    Pi          -
    TwoPI       -
    InvRad      - Radians per degree

    Coupling      :

    References             :
    1996 Astronomical Almanac Pg. C24
    http://aa.usno.navy.mil/faq/docs/SunApprox.php
    """
    N = jd - 2451545.0

    meanlong = 280.461 + 0.9856474 * N
    meanlong = attitude.normalize(meanlong, 0, 360)[0]

    meananomaly = 357.528 + 0.9856003 * N
    meananomaly = attitude.normalize(meananomaly * deg2rad, 0, twopi)[0]
    if meananomaly < 0:
        meananomaly = twopi + meananomaly

    eclplong = meanlong + 1.915 * \
        np.sin(meananomaly) + 0.020 * np.sin(2 * meananomaly)
    obliquity = 23.439 - 0.0000004 * N

    meanlong = meanlong * deg2rad
    if meanlong < 0:
        meanlong = twopi + meanlong

    eclplong = eclplong * deg2rad
    obliquity = obliquity * deg2rad

    ra = np.arctan2(np.cos(obliquity) * np.sin(eclplong), np.cos(eclplong))
    dec = np.arcsin(np.sin(obliquity) * np.sin(eclplong))

    # equation of time
    eqtime = meanlong * rad2deg / 15 - ra * rad2deg / 15

    # sun vector
    sun_dist = 1.00014 - 0.01671 * \
        np.cos(meananomaly) - 0.00014 * np.cos(2 * meananomaly)

    semidiameter = 0.2666 / sun_dist  # angular semidiamter in deg
    sun_eci = [np.cos(eclplong) * sun_dist * au2km,
               np.cos(obliquity) * np.sin(eclplong) * sun_dist * au2km,
               np.sin(obliquity) * np.sin(eclplong) * sun_dist * au2km]

    return np.squeeze(sun_eci), ra, dec


def planet_coe(JD_curr, planet_flag):
    r"""Orbital elements for any of the major solar system bodies

    Given the Julian date this function will output the approximate orbital
    elements for any of the solar system bodies. It uses an analytic
    approximation for their positions and propogates the orbital elements to
    the desired date.

    Parameters
    ----------
    jd : float
        Julian date for the epoch
    planet_flag : int
        Number 0 to 8 for each of the planets (Pluto forever)

    Returns
    -------
    coe : tuple
        p : float
            Semiparameter of orbit in AU
        ecc: float
            Eccentricity of orbit
        inc : float
            Inclination in radiands wrt to ecliptic
        raan : float
            Longitude/Right ascension of teh ascending node in radians
        argp : float
            Argument of perigee in radians
        nu : float
            True anomaly in radians
    r_ecliptic : (3,) ndarray
        Position wrt J2000 mean ecliptic (Earth's orbit around the sun) in
        kilometers defined relative to the solar system barycenter
    v_ecliptic : (3,) ndarray
        Velocity wrt J2000 mean ecliptic in kilometers per second defined
        relative to the solar system barycenter
    r_icrf : (3,) ndarray
        Position wrt J2000 equatorial/ICRF in kilometers and defined relative
        to teh solar system barycenter
    v_icrf : (3,)
        Velocity wrt J2000 equatorial/ICRF in kilometers per second defined
        relative to the solar system barycenter

    See Also
    --------
    planet_approx : Actually has the data for the approximations 

    Notes
    -----
    The orbital elements are defined with respect to the mean ecliptic and
    equinox of J2000. The fundamental plane is the ecliptic, or the orbit of
    the Earth about the sun. The fundamental direction is aligned with the
    first point of Ares or the vernal equinox.

    The vectors are defined in both the ecliptic and equatorial system with
    respect to the solar system barycenter

    Author
    ------
    Shankar Kulumani		GWU		skulumani@gwu.edu   23 June 2017

    References
    ----------

    .. [1] https://ssd.jpl.nasa.gov/?planet_pos
    .. [2] MEEUS, Jean H.. Astronomical Algorithms. Willmann-Bell,
    Incorporated, 1991.

    Examples
    --------
    An example of how to use the function

    >>> planet_flag = 4
    >>> coe = planet_coe(jd, planet_flag)

    """

    # Find the JD centuries past J2000 epoch
    JD_J2000 = 2451545.0
    T = (JD_curr - JD_J2000) / 36525

    # load the elements and element rates for the planets
    (a0, adot, e0, edot, inc0, incdot, meanL0, meanLdot, lonperi0, lonperidot,
     raan0, raandot, b, c, f, s) = planet_approx(JD_curr, planet_flag)

    # compute the current elements at this JD
    a = a0 + adot * T
    ecc = e0 + edot * T
    inc = inc0 + incdot * T
    L = meanL0 + meanLdot * T
    lonperi = lonperi0 + lonperidot * T
    raan = raan0 + raandot * T

    # compute argp and M/v to complete the element set
    argp = lonperi - raan
    M = L - lonperi + b * T**2 + c * np.cos(f * T) + s * np.sin(f * T)

    M = attitude.normalize(M, -180, 180)
    # solve kepler's equation to compute E and v
    E, nu, count = kepler.kepler_eq_E(np.deg2rad(M), ecc)
    
    # package into a vector and output
    p = a * (1 - ecc**2)
    
    coe = COE(p=p, ecc=ecc, inc=np.deg2rad(inc), raan=np.deg2rad(raan),
           argp=attitude.normalize(np.deg2rad(argp), 0, 2*np.pi), nu=nu)

    # convert to position and velocity vectors in J2000 ecliptic and J2000
    # Earth equatorial (ECI) reference frame
    # need to convert distances to working units of kilometers
    au2km = constants.au2km
    
    r_ecliptic, v_ecliptic, r_pqw, v_pqw = kepler.coe2rv(p * au2km, ecc, inc,
                                                         raan, argp, nu.item(),
                                                         constants.sun.mu)

    # convert to the Earth J2000 frame (ECI) need to rotate by the obliquity of
    # the ecliptic
    Eclip2ECI = attitude.rot1(constants.obliquity)

    r_eci = Eclip2ECI.dot(r_ecliptic)
    v_eci = Eclip2ECI.dot(v_ecliptic)

    return coe, r_ecliptic, v_ecliptic, r_eci, v_eci


def planet_approx(JD, planet_flag):
    r"""Approximate orbital elements of each of the planets

    This function provides the classical orbital elements and the rates of
    change for any of the planets in the range of 3000 BC to 3000 AD. A higher
    accuracy model if provided over the range of 1800 - 2050 AD.

    Parameters
    ----------
    jd : float
        Julian date for the epoch
    planet_flag : int
        Number 0 to 8 for each of the planets (Pluto forever)

    Returns
    -------
    a0 : float
        Semimajor axis in AU at epoch
    adot : float
        Rate of change of semimajor axis in AU/century
    e0 : float
        Eccentricty of orbit at epoch
    edot :  float
        Rate of change of eccentricy per century
    inc0 : float
        inclination in degrees at epoch
    incdot : float
        Rate of change of inclination in degrees/century
    meanL0 : float
        Mean longitude in degrees at epoch
    meanLdot : float
        Rate of change of mean longitude in degrees/century
    lonperi0 : float
        Longitude of perihelion in degrees at epoch
    lonperidot : float
        Rate of change of Longitude of perihelion in degrees/century
    raan0 : float
        Longitude of the ascending node in degrees at epoch
    raandot : float
        Rate of RAAN in degrees/century
    b, c, f, s : floats
        Additional parameters for the outer planets when using the coarse model

    See Also
    --------
    planet_coe : Use this function to compute the orbital elements

    Notes
    -----
    The orbital elements are defined with respect to the mean ecliptic and
    equinox of J2000. The fundamental plane is the ecliptic, or the orbit of
    the Earth about the sun. The fundamental direction is aligned with the
    first point of Ares or the vernal equinox.

    Rather than using this function directly, simply use planet_coe to get the
    classical orbital elements wrt J2000 ecliptic.

    Author
    ------
    Shankar Kulumani		GWU		skulumani@gwu.edu   23 June 2017

    References
    ----------

    .. [1] https://ssd.jpl.nasa.gov/?planet_pos
    .. [2] MEEUS, Jean H.. Astronomical Algorithms. Willmann-Bell,
    Incorporated, 1991.

    Examples
    --------
    An example of how to use the function


    """
    JD_1800AD = 2378497.000000
    JD_2050AD = 2469808.000000

    JD_3000BC = 625674.000000
    JD_3000AD = 2816788.000000

    if JD > JD_1800AD and JD < JD_2050AD:  # date is between 1800-2050AD and use the fine model
        b = 0.0
        c = 0.0
        f = 0.0
        s = 0.0

        if planet_flag == 0:  # mercury
            a0 = 0.38709927
            adot = 0.00000037

            e0 = 0.20563593
            edot = 0.00001906

            inc0 = 7.00497902
            incdot = -0.00594749

            meanL0 = 252.25032350
            meanLdot = 149472.67411175

            lonperi0 = 77.45779628
            lonperidot = 0.16047689

            raan0 = 48.33076593
            raandot = -0.12534081

        elif planet_flag == 1:  # venus
            a0 = 0.72333566
            adot = 0.00000390

            e0 = 0.00677672
            edot = -0.00004107

            inc0 = 3.39467605
            incdot = -0.00078890

            meanL0 = 181.97909950
            meanLdot = 58517.81538729

            lonperi0 = 131.60246718
            lonperidot = 0.00268329

            raan0 = 76.67984255
            raandot = -0.27769418
        elif planet_flag == 2:  # earth moon barycenter
            a0 = 1.00000261
            adot = 0.00000562

            e0 = 0.01671123
            edot = -0.00004392

            inc0 = -0.00001531
            incdot = -0.01294668

            meanL0 = 100.46457166
            meanLdot = 35999.37244981

            lonperi0 = 102.93768193
            lonperidot = 0.32327364

            raan0 = 0.0
            raandot = 0.0
        elif planet_flag == 3:  # mars
            a0 = 1.52371034
            adot = 0.00001847

            e0 = 0.09339410
            edot = 0.00007882

            inc0 = 1.84969142
            incdot = -0.00813131

            meanL0 = -4.55343205
            meanLdot = 19140.30268499

            lonperi0 = -23.94362959
            lonperidot = 0.44441088

            raan0 = 49.55953891
            raandot = -0.29257343
        elif planet_flag == 4:  # jupiter
            a0 = 5.20288700
            adot = -0.00011607

            e0 = 0.04838624
            edot = -0.00013253

            inc0 = 1.30439695
            incdot = -0.00183714

            meanL0 = 34.39644051
            meanLdot = 3034.74612775

            lonperi0 = 14.72847983
            lonperidot = 0.21252668

            raan0 = 100.47390909
            raandot = 0.20469106
        elif planet_flag == 5:  # saturn
            a0 = 9.53667594
            adot = -0.00125060

            e0 = 0.05386179
            edot = -0.00050991

            inc0 = 2.48599187
            incdot = 0.00193609

            meanL0 = 49.95424423
            meanLdot = 1222.49362201

            lonperi0 = 92.59887831
            lonperidot = -0.41897216

            raan0 = 113.66242448
            raandot = -0.28867794
        elif planet_flag == 6:  # uranus
            a0 = 19.18916464
            adot = -0.00196176

            e0 = 0.04725744
            edot = -0.00004397

            inc0 = 0.77263783
            incdot = -0.00242939

            meanL0 = 313.23810451
            meanLdot = 428.48202785

            lonperi0 = 170.95427630
            lonperidot = 0.40805281

            raan0 = 74.01692503
            raandot = 0.04240589
        elif planet_flag == 7:  # neptune
            a0 = 30.06992276
            adot = 0.00026291

            e0 = 0.00859048
            edot = 0.00005105

            inc0 = 1.77004347
            incdot = 0.00035372

            meanL0 = -55.12002969
            meanLdot = 218.45945325

            lonperi0 = 44.96476227
            lonperidot = -0.32241464

            raan0 = 131.78422574
            raandot = -0.00508664
        elif planet_flag == 8:  # pluto
            a0 = 39.48211675
            adot = -0.00031596

            e0 = 0.24882730
            edot = 0.00005170

            inc0 = 17.14001206
            incdot = 0.00004818

            meanL0 = 238.92903833
            meanLdot = 145.20780515

            lonperi0 = 224.06891629
            lonperidot = -0.04062942

            raan0 = 110.30393684
            raandot = -0.01183482
        else:
            print("Incorrect planet flag should be between 0 and 8")

    elif JD > JD_3000BC and JD < JD_3000AD:
        # we'll set this for the outer planets. everyone else gets zero
        b = 0.0
        c = 0.0
        f = 0.0
        s = 0.0

        if planet_flag == 0:  # mercury
            a0 = 0.38709843
            adot = 0.0

            e0 = 0.20563661
            edot = 0.00002123

            inc0 = 7.00559432
            incdot = -0.00590158

            meanL0 = 252.25166724
            meanLdot = 149472.67486623

            lonperi0 = 77.45771895
            lonperidot = 0.15940013

            raan0 = 48.33961819
            raandot = -0.12214182

        elif planet_flag == 1:  # venus
            a0 = 0.72332102
            adot = -0.00000026

            e0 = 0.00676399
            edot = -0.00005107

            inc0 = 3.39777545
            incdot = 0.00043494

            meanL0 = 181.97970850
            meanLdot = 58517.81560260

            lonperi0 = 131.76755713
            lonperidot = 0.05679648

            raan0 = 76.67261496
            raandot = -0.27274174

        elif planet_flag == 2:  # earth moon barycenter
            a0 = 1.00000018
            adot = -0.00000003

            e0 = 0.01673163
            edot = -0.00003661

            inc0 = -0.00054346
            incdot = -0.01337178

            meanL0 = 100.46691572
            meanLdot = 35999.37306329

            lonperi0 = 102.93005885
            lonperidot = 0.31795260

            raan0 = -5.11260389
            raandot = -0.24123856
        elif planet_flag == 3:  # mars
            a0 = 1.52371243
            adot = 0.00000097

            e0 = 0.09336511
            edot = 0.00009149

            inc0 = 1.85181869
            incdot = -0.00724757

            meanL0 = -4.56813164
            meanLdot = 19140.29934243

            lonperi0 = -23.91744784
            lonperidot = 0.45223625

            raan0 = 49.71320984
            raandot = -0.26852431
        elif planet_flag == 4:  # jupiter
            a0 = 5.20248019
            adot = -0.00002864

            e0 = 0.04853590
            edot = 0.00018026

            inc0 = 1.29861416
            incdot = -0.00322699

            meanL0 = 34.33479152
            meanLdot = 3034.90371757

            lonperi0 = 14.27495244
            lonperidot = 0.18199196

            raan0 = 100.29282654
            raandot = 0.13024619
        elif planet_flag == 5:  # saturn
            a0 = 9.54149883
            adot = -0.00003065

            e0 = 0.05550825
            edot = -0.00032044

            inc0 = 2.49424102
            incdot = 0.00451969

            meanL0 = 50.07571329
            meanLdot = 1222.11494724

            lonperi0 = 92.86136063
            lonperidot = 0.54179478

            raan0 = 113.63998702
            raandot = -0.25015002
        elif planet_flag == 6:  # uranus
            a0 = 19.18797948
            adot = -0.00020455

            e0 = 0.04685740
            edot = -0.00001550

            inc0 = 0.77298127
            incdot = -0.00180155

            meanL0 = 314.20276625
            meanLdot = 428.49512595

            lonperi0 = 172.43404441
            lonperidot = 0.09266985

            raan0 = 73.96250215
            raandot = 0.05739699
        elif planet_flag == 7:  # neptune
            a0 = 30.06952752
            adot = 0.00006447

            e0 = 0.00895439
            edot = 0.00000818

            inc0 = 1.77005520
            incdot = 0.00022400

            meanL0 = 304.22289287
            meanLdot = 218.46515314

            lonperi0 = 46.68158724
            lonperidot = 0.01009938

            raan0 = 131.78635853
            raandot = -0.00606302
        elif planet_flag == 8:  # pluto
            a0 = 39.48686035
            adot = 0.00449751

            e0 = 0.24885238
            edot = 0.00006016

            inc0 = 17.14104260
            incdot = 0.00000501

            meanL0 = 238.96535011
            meanLdot = 145.18042903

            lonperi0 = 224.09702598
            lonperidot = -0.00968827

            raan0 = 110.30167986
            raandot = -0.00809981
        else:
            print("Incorrect planet flag should be between 0 and 8")
    else:
        raise ValueError('Julian Date is outside of the range 3000BC-3000AD')

    return (a0, adot, e0, edot, inc0, incdot, meanL0, meanLdot, lonperi0,
            lonperidot, raan0, raandot, b, c, f, s)


def asteroid_epoch(ast_flag):
    """
    This holds the orbital elements for the asteroids as taken from JPL

    Return the state at the JD epoch for use in a propogate function
    """

    if ast_flag == 0:
        # 2008 EV5
        a = 0.9582899238313918
        ecc = 0.08348599378460778
        p = a * (1 - ecc**2)
        inc = np.deg2rad(7.436787362690259)
        raan = np.deg2rad(93.39122898787916)
        argp = np.deg2rad(234.8245876826614)
        M = np.deg2rad(3.409187469072454)
        JD_epoch = 2457800.5
    elif ast_flag == 1:
        # Itokawa
        a = 1.324163617639197
        ecc = 0.28011765678781
        p = a * (1 - ecc**2)
        inc = np.deg2rad(1.62145641293925)
        raan = np.deg2rad(69.07992986350325)
        argp = np.deg2rad(162.8034822691509)
        M = np.deg2rad(131.4340297670125)
        JD_epoch = 2457800.5
    elif ast_flag == 2:
        # Bennu
        a = 1.126391026007489
        ecc = 0.2037451112033579
        p = a * (1 - ecc**2)
        inc = np.deg2rad(6.034939195483961)
        raan = np.deg2rad(2.060867837066797)
        argp = np.deg2rad(66.22306857848962)
        M = np.deg2rad(101.7039476994243)
        JD_epoch = 2455562.5
    else:
        print("No such asteroid defined yet. Using Itokawa so nothing breaks instead")

        # Itokawa
        a = 1.324163617639197
        ecc = 0.28011765678781
        p = a * (1 - ecc**2)
        inc = np.deg2rad(1.62145641293925)
        raan = np.deg2rad(69.07992986350325)
        argp = np.deg2rad(162.8034822691509)
        M = np.deg2rad(131.4340297670125)
        JD_epoch = 2457800.5

    # convert M to nu for later and output the coe
    E, nu, count = kepler_eq_E(M, ecc)

    epoch = (p, ecc, inc, raan, argp, attitude.normalize(nu, 0, 2 * np.pi), JD_epoch)

    return epoch


def asteroid_coe(JD_curr, ast_flag):
    """
        Output the current COE for the chosen asteroid
    """

    # load the asteroid COE at the epoch
    (p, ecc, inc, raan, argp, nu_0, JD_epoch) = asteroid_epoch(ast_flag)

    # compute the delta t
    delta_t = (JD_curr - JD_epoch) * 86400
    mu = 1.32712440018e20  # m^3 / s^2
    mu = 1 / 149597870700**3 * mu  # au^3 / sec^2

    # propogate to the current JD_curr
    (E_f, M_f, nu_f) = tof_delta_t(p, ecc, mu, nu_0, delta_t)

    # output the current COE
    coe = (p, ecc, inc, raan, argp, attitude.normalize(nu_f, 0, 2 * np.pi))

    return coe



"""Astrodynamic constants

This module will load a wide variety of WGS84 derived constants as module
attributes. These constants may be referenced in subsequent scripts and functions.
These variables are case specific and must be noted.

Notes
-----

To use these constants in your scripts/functions, simply import this module

from astro import constants

Then use any of the attributes as:

deg2rad = constants.deg2rad

Attributes
----------
constants : module level attributes broken down by planetary body
    rot_per : axial rotational period in rev/day
    radius : mean equatorial radius in km
    mu : graviational parameter in km^3/sec^2
    orbit_sma : semi-major axis of orbit in km
    orbit_per : period of orbit in sec
    orbit_ecc : eccentricity of orbit
    orbit_inc : inclination of orbit in deg
    mass : body mass (derived from mu and G) in kg
    G : universal graviational constant in km^3/kg*sec^2

Author
------
Shankar Kulumani 1 Sept 2012
    list revisions
Shankar Kulumani 19 Sept 2012
    added distance conversions
    added time conversions
Shankar Kulumani 30 Oct 2012
    added semiparamter to constants
Shankar Kulumani 21 June 2017
    Created for MAE3145 in Python

References
----------
Astro 321 USAFA 2007
AAE 532 Purdue 2012 planetary_constants.pdf (from AAE532 supplements)

"""
from __future__ import absolute_import, division, print_function, unicode_literals
from collections import namedtuple
import numpy as np

# HALFPI,PI2           PI/2, & 2PI in various names
halfpi = np.pi / 2.0
twopi = 2.0 * np.pi

zero_inc_ecc = 0.015                # Small number for incl & ecc purposes
small = 1.0E-6                  # Small number used for tolerance purposes
undefined = 999999.1

# General constants
deg2rad = np.pi / 180.0
rad2deg = 180.0 / np.pi

G = 6.673e-20
au2km = 149597870.0  # km
km2au = 1 / au2km
km2re = 1 / 6378.137  # km to Earth radii
re2km = 6378.137

sec2hr = 1 / 3600
hr2sec = 3600
day2sec = 86400.0                   # sec/day
sec2day = 1 / 86400

# Earth Characteristics from WGS 84
obliquity = np.radians(23.43928)

#  Moon & Sun Characteristics from WGS 84
mu_moon = 4902.774191985            # km^3/sec^2
mu_sun = 1.32712438e11              # km^3/sec^2

# Planetary constants
# namedtuple to hold constants for each body
Body = namedtuple('Body', ['rot_per', 'radius', 'mu', 'mass', 'orbit_sma',
                           'orbit_per', 'orbit_ecc', 'orbit_inc', 'p'])

Earth = namedtuple('Earth', ['rot_per', 'radius', 'mu', 'mass', 'orbit_sma',
                             'orbit_per', 'orbit_ecc', 'orbit_inc', 'p',
                             'omega', 'sidepersol', 'radperday', 'flat',
                             'eesqrd', 'ee', 'J2', 'J3', 'J4'])
# earth
earth = Earth(rot_per=1.0027576,
              radius=6378.137,
              mu=398600.5,
              orbit_sma=1.49589800e8,
              orbit_per=31555647.16,
              orbit_ecc=.01671020,
              orbit_inc=4.98816000e-5,
              mass=398600.5 / G,
              p=1.49589800e8 * (1 - .01671020**2),
              omega=0.000072921151467,     # rad/sec
              sidepersol=1.00273790935,          # Sidereal Days/Solar Day
              radperday=6.30038809866574,        # rad/day for the Earth
              flat=1.0 / 298.257223563,            # Earth flattening factor
              # Earth eccentricty squared
              eesqrd=(2.0 - (1.0 / 298.257223563)) * (1.0 / 298.257223563),
              ee=8.1819190842622e-2,
              J2=0.00108263,                     # Zonal Harmonic perturbation
              J3=-0.00000254,
              J4=-0.00000161)

# define sun constants
sun = Body(rot_per=0.0394011,
           radius=695990.0,
           mu=1.32712000e11,
           mass=1.32712000e11 / G,
           orbit_sma=0,
           orbit_per=0,
           orbit_ecc=0,
           orbit_inc=0,
           p=0)


# define moon constants
moon = Body(rot_per=.0366004,
            radius=1737.50,
            mu=4.902800e3,
            orbit_sma=3.84400000e5,
            orbit_per=2360592,
            orbit_ecc=.05540000,
            orbit_inc=5.16000,
            mass=4.902800e3 / G,
            p=3.84400000e5 * (1 - .05540000**2))

# mercury
mercury = Body(rot_per=.0170515,
               radius=2.439000e3,
               mu=2.203210e4,
               orbit_sma=5.79092000e7,
               orbit_per=7600568.601,
               orbit_ecc=0.205631,
               orbit_inc=7.00487,
               mass=2.203210e4 / G,
               p=5.79092000e7 * (1 - 0.205631**2))
# venus
venus = Body(rot_per=.0041149,
             radius=6051.80,
             mu=3.24859000e5,
             orbit_sma=1.0820900e8,
             orbit_per=19414191.77,
             orbit_ecc=.006773,
             orbit_inc=3.39471,
             mass=3.24859000e5 / G,
             p=1.0820900e8 * (1 - .006773**2))

# mars
mars = Body(
    rot_per=.9746985,
    radius=3397.00,
    mu=4.282840e4,
    orbit_sma=2.27937000e8,
    orbit_per=59353583.28,
    orbit_ecc=.09341230,
    orbit_inc=1.85061,
    mass=4.282840e4 / G,
    p=2.27937000e8 * (1 - .09341230**2))

# jupiter
jupiter = Body(
    rot_per=2.4181458,
    radius=71492.00,
    mu=1.26687000e8,
    orbit_sma=7.7841200e8,
    orbit_per=374396573,
    orbit_ecc=.04839270,
    orbit_inc=1.30530,
    mass=1.26687000e8 / G,
    p=7.7841200e8 * (1 - .04839270**2))

callisto = Body(
    rot_per=0,
    radius=0,
    mu=7179.29,
    orbit_sma=1.883e6,
    orbit_per=0,
    orbit_ecc=0,
    orbit_inc=0,
    mass=0,
    p=0)

#  saturn
saturn = Body(
    rot_per=2.2522523,
    radius=60330.00,
    mu=3.79313000e7,
    orbit_sma=1.42673000e9,
    orbit_per=929341659.8,
    orbit_ecc=.05415060,
    orbit_inc=2.48446,
    mass=3.79313000e7 / G,
    p=1.42673000e9 * (1 - .05415060**2))


titan = Body(
    rot_per=0.627393,
    radius=2575.5,
    mu=8978.19,
    orbit_sma=1221830.0,
    orbit_per=15.94542*day2sec,
    orbit_ecc=0.0292,
    orbit_inc=0.33,
    mass=8978.19 / G,
    p=1221830.0 * (1 - 0.2**2))

# uranus
uranus = Body(
    rot_per=1.3921178,  # retrograde rotation
    radius=26200.00,
    mu=5.79397000e6,
    orbit_sma=2.87097000e9,
    orbit_per=2653128427,
    orbit_ecc=.04716770,
    orbit_inc=.76986,
    mass=5.79397000e6 / G,
    p=2.87097000e9 * (1 - .04716770**2))

# neptune
neptune = Body(
    rot_per=1.4897579,
    radius=25225.00,
    mu=6.835110e6,
    orbit_sma=4.49825000e9,
    orbit_per=5203301252,
    orbit_ecc=.00858587,
    orbit_inc=1.76917,
    mass=6.835110e6 / G,
    p=4.49825000e9 * (1 - .00858587**2))

# pluto
pluto = Body(
    rot_per=.1565631,
    radius=1195.00,
    mu=8.737670e2,
    orbit_sma=5.906638e9,
    orbit_per=7829522968,
    orbit_ecc=.24880800,
    orbit_inc=17.14180,
    mass=5.906638e9 / G,
    p=5.906638e9 * (1 - .24880800**2))



"""Keplerian method to do astrodynamics
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
from kinematics import attitude
import pdb
from . import constants


def coe2rv(p_in, ecc_in, inc_in, raan_in, arg_p_in, nu_in, mu):
    """
       Purpose:
           - Convert the classical orbital elements (COEs) to position (R)
               and velocity (V) vectors.

       [R_ijk,V_ijk,R_pqw,V_pqw] =
       coe2rv(p,ecc,inc,raan,arg_p,nu,mu,arglat,truelon,lonper )

       Inputs:
           - p - semi-major axis (km)
           - ecc - eccentricity
           - raan - right acsension of the ascending node (rad) 0 < raan <2*pi
           - inc - inclination (rad) 0 < inc < pi
           - arg_p - argument of periapsis (rad) 0 < arg_p < 2*pi
           - nu - true anomaly (rad) 0 < nu < 2*pi
           - mu - gravitational parameter of central body (km^3/sec^2).
           - arglat - argument of latitude(CI) rad 0 <arglat < 2*pi
           - truelon - true longitude (CE) rad 0 < truelon < 2*pi
           - lonper - longitude of periapsis rad 0 < lonper < 2*pi

       Outpus:
           - R_ijk - position vector in inertial frame (km)
           - V_ijk - velocity vector in inertial frame (km/sec)
           - R_pqw - position vecotr in perifocal frame (km)
           - V_pqw - velocity vector in perifocal frame (km/sec)

       Dependencies:
           - ROT1 - elementary rotation about first axis
           - ROT2 - elementary rotation about second axis
           - ROT3 - elementary rotation about third axis

       Author:
           - Shankar Kulumani 18 Aug 2012
               - revised from code written at USAFA Fall 2007
           - Shankar Kulumani 18 Sept 2012
               - modified rotation matrix from PQW to ECI
           - Shankar Kulumani 29 Sept 2012
               - modified input to take semi-latus rectum to allow calculation
               for all orbit types
           - Shankar Kulumani 7 Dec 2014
               - loop for vector inputs
           - Shankar Kulumani 7 Dec 2016
               - Convert to python and remove vectorized inputs

    """

    tol = 1e-9
    # check if array inputs
    if not hasattr(p_in, "__iter__"):
        p_in = np.asarray([p_in], dtype=np.float)
        ecc_in = np.asarray([ecc_in], dtype=np.float)
        inc_in = np.asarray([inc_in], dtype=np.float)
        raan_in = np.asarray([raan_in], dtype=np.float)
        arg_p_in = np.asarray([arg_p_in], dtype=np.float)
        nu_in = np.asarray([nu_in], dtype=np.float)
    # make sure all inputs are the same size
    if not (p_in.shape == ecc_in.shape == inc_in.shape == raan_in.shape ==
            arg_p_in.shape == nu_in.shape):
        print("All inputs must be the same shape")
        print("p: {}".format(p_in.shape))
        print("ecc: {}".format(ecc_in.shape))
        print("inc: {}".format(inc_in.shape))
        print("raan: {}".format(raan_in.shape))
        print("argp: {}".format(arg_p_in.shape))
        print("nu: {}".format(nu_in.shape))

    r_pqw_out = []
    v_pqw_out = []
    r_ijk_out = []
    v_ijk_out = []

    for p, ecc, inc, raan, arg_p, nu in zip(p_in, ecc_in, inc_in, raan_in,
                                            arg_p_in, nu_in):
        # check eccentricity for special cases
        if (ecc < tol):
            # circular equatorial
            if (inc < tol) or (abs(inc - np.pi) < tol):
                arg_p = 0.0
                raan = 0.0
                # nu   = truelon
                nu = raan + arg_p + nu
            else:
                # circular inclined
                arg_p = 0.0
                nu = arg_p + nu
                # nu = arglat
        elif ((inc < tol) or (abs(inc - np.pi) < tol)):    # elliptical equatorial
            arg_p = raan + arg_p
            raan = 0.0
            # arg_p=lonper

        cosnu = np.cos(nu)
        sinnu = np.sin(nu)
        radius = p / (1 + ecc * cosnu)
        # semi latus rectum check
        if (abs(p) < 0.0001):
            p = 0.0001

        # calculate postion and velocity in perifocal frame
        r_pqw = radius * np.array([cosnu, sinnu, 0])
        v_pqw = np.sqrt(mu / p) * np.array([-sinnu, (ecc + cosnu), 0])

        PI = attitude.rot3(raan).dot(
            attitude.rot1(inc)).dot(attitude.rot3(arg_p))

        # rotate postion and velocity vectors to inertial frame
        r_ijk = np.dot(PI, r_pqw)
        v_ijk = np.dot(PI, v_pqw)

        r_pqw_out.append(r_pqw)
        v_pqw_out.append(v_pqw)
        r_ijk_out.append(r_ijk)
        v_ijk_out.append(v_ijk)

    return (np.squeeze(r_ijk_out), np.squeeze(v_ijk_out), np.squeeze(r_pqw_out),
            np.squeeze(v_pqw_out))

def hne_vec(r, v, mu):
    r"""Compute fundamental vectors associated with orbit

    This will compute the angular momentum, h, nodal, n, and eccentricity
    vector, e, for a Keplerian two-body orbit.

    Parameters
    ----------
    r : array_like and type
        <`4:Description of the variable`>

    Returns
    -------
    describe : type
        Explanation of return value named describe

    Other Parameters
    ----------------
    only_seldom_used_keywords : type
        Explanation of this parameter

    Raises
    ------
    BadException
        Because you shouldn't have done that.

    See Also
    --------
    other_func: Other function that this one might call

    Notes
    -----
    You may include some math:

    .. math:: X(e^{j\omega } ) = x(n)e^{ - j\omega n}

    Author
    ------
    Shankar Kulumani		GWU		skulumani@gwu.edu

    References
    ----------
    Cite the relevant literature, e.g. [1]_.  You may also cite these
    references in the notes section above.

    .. [1] Shannon, Claude E. "Communication theory of secrecy systems."
    Bell Labs Technical Journal 28.4 (1949): 656-715

    Examples
    --------
    An example of how to use the function

    >>> a = [1, 2, 3]
    >>> print [x + 3 for x in a]
    [4, 5, 6]
    >>> print "a\n\nb"
    a
    b

    """ 
    
    # compute angular momentum vector
    mag_r = np.linalg.norm(r)
    mag_v = np.linalg.norm(v)
    rdotv = r.dot(v)

    h = np.cross(r, v)
    mag_h = np.linalg.norm(h)

    h_hat = h / mag_h

    # compute line of nodes vector
    n = np.zeros(3)
    n[0] = -h[1]
    n[1] = h[0]
    n[2] = 0.0
    
    if np.linalg.norm(n) < constants.small:
        mag_n = 0
        n_hat = np.zeros(3)
    else:
        mag_n = np.linalg.norm(n)
        n_hat = n / mag_n

    # compute eccentricity vector
    e = ((mag_v**2 - mu / mag_r) * r - rdotv * v) / mu
    if np.linalg.norm(e) < constants.small:
        e_hat = np.zeros(3)
    else:
        e_hat = e / np.linalg.norm(e)
    
    return h_hat, n_hat, e_hat

def rv2coe(r, v, mu):
    """Position and Velocity vectors to classical orbital elements

    Purpose:
        - Converts inertial position and velocity vectors to orbital
        elements

            [p,a,ecc,inc,raan,arg_p,nu,m,arglat,truelon,lonper] = rv2coe (r,v, mu)

    Inputs:
        - r - position vector in inertial frame (km)
        - v - velocity vector in inertial frame (km/sec)
        - mu - gravitational parameter of central body (km^3/sec^2)

    Outputs:
        - p - semi-major axis (km)
        - ecc - eccentricity
        - raan - right acsension of the ascending node (rad) 0 < raan <
        2*pi
        - inc - inclination (rad) 0 < inc < pi
        - arg_p - argument of periapsis (rad) 0 < arg_p < 2*pi
        - nu - true anomaly (rad) 0 < nu < 2*pi
        - m - mean anomaly in rad
        - arglat - argument of latitude(CI) rad 0 <arglat < 2*pi
        - truelon - true longitude (CE) rad 0 < truelon < 2*pi
        - lonper - longitude of periapsis rad 0 < lonper < 2*pi

    Dependencies:
        - None

    Author:
        - Shankar Kulumani 30 Sept 2012
            - used old USAFA code and AAE532
        - Shankar Kulumani 5 September 2017
            - convert to Python for use in MAE3145

    References
        - AAE532 Notes
        - Vallado 3rd Edition
    """

    tol = 1e-9

    # calculate angular moemntum vector and magnitude
    mag_r = np.linalg.norm(r)
    mag_v = np.linalg.norm(v)

    rdotv = np.dot(r, v)

    h = np.cross(r, v)
    mag_h = np.linalg.norm(h)

    h_hat = h / mag_h
    # find n,e
    if mag_h > tol:
        n = np.zeros(3)
        n[0] = -h[1]
        n[1] = h[0]
        n[2] = 0.0

        with np.errstate(divide='raise'):
            try:
                mag_n = np.linalg.norm(n)
                n_hat = n / mag_n
            except FloatingPointError:
                mag_n = 0
                n_hat = np.zeros(3)

        # eccentricity vector
        e = ((mag_v**2 - mu / mag_r) * r - rdotv * v) / mu

        ecc = np.linalg.norm(e)
        mag_e = ecc

        sme = mag_v**2 / 2 - mu / mag_r

        if np.absolute(sme) > tol:
            a = -mu / (2 * sme)
        else:
            a = np.inf

        p = mag_h**2 / mu

        # inclination
        inc = np.arccos(h[2] / mag_h)

        # determine orbit type
        orbit_type = 'ei'
        if ecc < tol:
            # circular equatorial
            if inc < tol or np.absolute(inc - np.pi) < tol:
                orbit_type = 'ce'
            else:
                # circular inclined
                orbit_type = 'ci'

        else:
            #  elliptical, parabolic, hyperbolic equatorial -
            if inc < tol or np.absolute(inc - np.pi) < tol:
                orbit_type = 'ee'

        # right ascension of the ascending node
        if mag_n > tol:
            temp = n[0] / mag_n
            if np.absolute(temp) > 1.0:
                temp = np.sign(temp)
            raan = np.arccos(temp)
            if n[1] < 0.0:
                raan = 2 * np.pi - raan
        else:
            raan = 0

        # find argument of periapsis
        if orbit_type == 'ei':
            arg_p = np.arccos(np.dot(n, e) / (mag_n * mag_e))
            if e[2] < 0.0:
                arg_p = 2 * np.pi - arg_p
        else:
            arg_p = 0

        # ------------ find true anomaly at epoch - ------------
        if orbit_type[0] == 'e':
            nu = np.arccos(np.dot(e, r) / (mag_e * mag_r))
            if rdotv < 0.0:
                nu = 2 * np.pi - nu
        else:
            nu = 0

        # special orbit cases
        # ---- find argument of latitude - circular inclined - ----
        if orbit_type == 'ci':
            arglat = np.arccos(np.dot(n, r) / (mag_n * mag_r))
            if r[2] < 0.0:
                arglat = 2 * pi - arglat
            m = arglat
            nu = arglat
        else:
            arglat = 0

        # -- find longitude of perigee - elliptical equatorial - ---
        if ecc > tol and orbit_type == 'ee':
            temp = e[0] / ecc
            if np.absolute(temp) > 1.0:
                temp = np.sign(temp)
            lonper = np.arccos(temp)
            if e[1] < 0.0:
                lonper = 2 * pi - lonper
            if inc > pi / 2:
                lonper = 2 * pi - lonper

            arg_p = lonper
        else:
            lonper = 0

        # -------- find true longitude - circular equatorial - -----
        if mag_r > tol and orbit_type == 'ce':
            temp = r[0] / mag_r
            if np.absolute(temp) > 1.0:
                temp = np.sign(temp)
            truelon = np.arccos(temp)
            if r[1] < 0.0:
                truelon = 2 * np.pi - truelon
            if inc > np.pi / 2:
                truelon = 2 * np.pi - truelon
            m = truelon

            nu = truelon
        else:
            truelon = 0

        # find mean anomaly
        if orbit_type[0] == 'e':
            E, m = nu2anom(nu, ecc)
    else:
        p = 0
        a = 0
        ecc = 0
        inc = 0
        raan = 0
        arg_p = 0
        nu = 0
        m = 0
        arglat = 0
        truelon = 0
        lonper = 0

    return p, a, ecc, inc, raan, arg_p, nu, m, arglat, truelon, lonper


def kepler_eq_E(M_in, ecc_in):
    """
    (E,nu,count) = kepler_eq_E(M,ecc)
    Purpose:
       - This function solves Kepler's equation for eccentric anomaly
       given a mean anomaly using a newton-rapson method.
           - Will work for elliptical/parabolic/hyperbolic orbits

    Inputs:
       - M - mean anomaly in rad -2*pi < M < 2*pi
       - ecc - eccentricity 0 < ecc < inf

    Outputs:
       - E - eccentric anomaly in rad 0 < E < 2*pi
       - nu - true anomaly in rad 0 < nu < 2*pi

    Dependencies:
       - none

    Author:
       - Shankar Kulumani 15 Sept 2012
           - rewritten from code from USAFA
           - solve for elliptical orbits add others later
       - Shankar Kulumani 29 Sept 2012
           - added parabolic/hyperbolic functionality
       - Shankar Kulumani 7 Dec 2014
          - added loop for vector inputs
       - Shankar Kulumani 2 Dec 2016
          - converted to python and removed the vector inputs

    References
       - USAFA Astro 321 LSN 24-25
       - Vallado 3rd Ed pg 72
    """
    tol = 1e-9
    max_iter = 50

    E_out = []
    nu_out = []
    count_out = []

    if not hasattr(M_in, "__iter__"):
        M_in = [M_in]

    if not hasattr(ecc_in, "__iter__"):
        ecc_in = [ecc_in]

    for M, ecc in zip(M_in, ecc_in):
        # eccentricity check
        """
            HYPERBOLIC ORBIT
        """
        if ecc - 1.0 > tol:  # eccentricity logic
            # initial guess
            if ecc < 1.6:  # initial guess logic
                if M < 0.0 and (M > -np.pi or M > np.pi):
                    E_0 = M - ecc
                else:
                    E_0 = M + ecc

            else:
                if ecc < 3.6 and np.absolute(M) > np.pi:
                    E_0 = M - np.sign(M) * ecc
                else:
                    E_0 = M / (ecc - 1.0)

            # netwon's method iteration to find hyperbolic anomaly
            count = 1
            E_1 = E_0 + ((M - ecc * np.sinh(E_0) + E_0) /
                         (ecc * np.cosh(E_0) - 1.0))
            while ((np.absolute(E_1 - E_0) > tol) and (count <= max_iter)):
                E_0 = E_1
                E_1 = E_0 + ((M - ecc * np.sinh(E_0) + E_0) /
                             (ecc * np.cosh(E_0) - 1.0))
                count = count + 1

            E = E_0
            # find true anomaly
            sinv = -(np.sqrt(ecc * ecc - 1.0) * np.sinh(E_1)) / \
                (1.0 - ecc * np.cosh(E_1))
            cosv = (np.cosh(E_1) - ecc) / (1.0 - ecc * np.cosh(E_1))
            nu = np.arctan2(sinv, cosv)
        else:
            """
                PARABOLIC
            """
            if np.absolute(ecc - 1.0) < tol:  # parabolic logic
                count = 1

                S = 0.5 * (np.pi / 2 - np.arctan(1.5 * M))
                W = np.arctan(np.tan(S)**(1.0 / 3.0))

                E = 2.0 * 1.0 / np.tan(2.0 * W)

                nu = 2.0 * np.arctan(E)
            else:
                """
                    ELLIPTICAl
                """
                if ecc > tol:   # elliptical logic

                    # determine intial guess for iteration
                    if M > -np.pi and (M < 0 or M > np.pi):
                        E_0 = M - ecc
                    else:
                        E_0 = M + ecc

                    # newton's method iteration to find eccentric anomaly
                    count = 1
                    E_1 = E_0 + (M - E_0 + ecc * np.sin(E_0)) / \
                        (1.0 - ecc * np.cos(E_0))
                    while ((np.absolute(E_1 - E_0) > tol) and (count <= max_iter)):
                        count = count + 1
                        E_0 = E_1
                        E_1 = E_0 + (M - E_0 + ecc * np.sin(E_0)) / \
                            (1.0 - ecc * np.cos(E_0))

                    E = E_0

                    # find true anomaly
                    sinv = (np.sqrt(1.0 - ecc * ecc) * np.sin(E_1)) / \
                        (1.0 - ecc * np.cos(E_1))
                    cosv = (np.cos(E_1) - ecc) / (1.0 - ecc * np.cos(E_1))
                    nu = np.arctan2(sinv, cosv)
                else:
                    """
                        CIRCULAR
                    """
                    # -------------------- circular -------------------
                    count = 0
                    nu = M
                    E = M

        E_out.append(E)
        nu_out.append(attitude.normalize(nu, 0, 2 * np.pi))
        count_out.append(count)

    return (np.squeeze(E_out), np.squeeze(nu_out), np.squeeze(count_out))


def conic_orbit(p, ecc, inc, raan, arg_p, nu_i, nu_f):
    """Plot conic orbit

        Purpose:
           - Uses the polar conic equation to plot a conic orbit

        [x y z xs ys zs ] = conic_orbit(p,ecc,inc,raan,arg_p,nu_i,nu_f)

        Inputs:
           - p - semi-major axis (km)
           - ecc - eccentricity
           - raan - right acsension of the ascending node (rad) 0 < raan <
           2*pi
           - inc - inclination (rad) 0 < inc < pi
           - arg_p - argument of periapsis (rad) 0 < arg_p < 2*pi
           - nu_i - initial true anomaly (rad) 0 < nu < 2*pi
           - nu_f - final true anomaly (rad) 0 < nu < 2*pi

        Outputs:
            pos_eci : (1000, 3) numpy array of satellite orbit in inertial frame
            sat_eci : (3,) numpy array of satellite position in inertial frame
            pos_pqw : (1000, 3) numpy array of orbit in perifocal frame
            sat_pqw : (3,) numpy array of satellite position in perifocal frame

        Dependencies:
           - ROT1,ROT2,ROT3 - principle axis rotation matrices

        Author:
           - Shankar Kulumani 1 Dec 2012
               - list revisions
           - Shankar Kulumani 5 Dec 2014
               - added outputs for orbit gui functions

        References
           - AAE532
    """

    tol = 1e-9
    step = 1000

    # v = true anomaly
    if nu_f > nu_i:
        v = np.linspace(nu_i, nu_f, step)
    else:
        v = np.linspace(nu_i, nu_f + 2 * np.pi, step)

    if ecc - 1 > tol:  # hyperbolic
        turn_angle = np.arccos(-1.0 / ecc)
        v = np.linspace(-turn_angle + 0.01, turn_angle - 0.01, step)

        if nu_i > np.pi:
            nu_i = nu_i - 2 * np.pi

        r = p / (1 + ecc * np.cos(v))
        rs = p / (1 + ecc * np.cos(nu_i))

    elif np.absolute(ecc - 1) < tol:  # parabolic
        v = np.linspace(-np.pi + 0.1, np.pi - 0.1, step)
        if nu_i > np.pi:
            nu_i = nu_i - 2 * np.pi

        r = p / 2 * (1 + np.tan(v / 2)**2)
        rs = p / 2 * (1 + np.tan(nu_i / 2)**2)
    else:
        # conic equation for elliptical orbit
        r = p / (1 + ecc * np.cos(v))
        rs = p / (1 + ecc * np.cos(nu_i))

    x = r * np.cos(v)
    y = r * np.sin(v)
    z = np.zeros_like(x)

    xs = rs * np.cos(nu_i)
    ys = rs * np.sin(nu_i)
    zs = 0

    pos_pqw = np.stack(( x, y, z ), axis=1)
    sat_pqw = np.array([ xs, ys, zs ])
    # rotate orbit plane to correct orientation

    # M_rot = [cos(raan) * cos(arg_p) - sin(raan) * cos(inc) * sin(arg_p) -cos(raan) * sin(arg_p) - sin(raan) * cos(inc) * cos(arg_p) sin(raan) * sin(inc);
    #         sin(raan) * cos(arg_p) + cos(raan) * cos(inc) * sin(arg_p) -sin(raan) * sin(arg_p) + cos(raan) * cos(inc) * cos(arg_p) -cos(raan) * sin(inc);
    #         sin(inc) * sin(arg_p) sin(inc) * cos(arg_p) cos(inc);];
    dcm_pqw2eci = attitude.rot3(-raan, 'r').dot(attitude.rot1(-inc, 'r')
                                           ).dot(attitude.rot3(-arg_p, 'r'))

    pos_eci = np.dot(dcm_pqw2eci, pos_pqw.T).T
    sat_eci = np.dot(dcm_pqw2eci, sat_pqw)

    return (pos_eci, sat_eci, pos_pqw, sat_pqw)


def nu2anom(nu, ecc):
    """
    [E M] = ecc_anomaly(nu,ecc)

       Purpose:
           - Calculates the eccentric and mean anomaly given eccentricity and
           true anomaly

       Inputs:
           - nu - true anomaly in rad -2*pi < nu < 2*pi
           - ecc - eccentricity of orbit 0 < ecc < inf

       Outputs:
           - E - (elliptical/parabolic/hyperbolic) eccentric anomaly in rad
               0 < E < 2*pi
           - M - mean anomaly in rad 0 < M < 2*pi

       Dependencies:
           - none

       Author:
           - Shankar Kulumani 5 Dec 2016
                - Convert to python
           - Shankar Kulumani 15 Sept 2012
               - modified from USAFA code and notes from AAE532
               - only elliptical case will add other later
           - Shankar Kulumani 17 Sept 2012
               - added rev check to reduce angle btwn 0 and 2*pi

       References
           - AAE532 notes
           - Vallado 3rd Ed
    """

    small = 1e-9

    if ecc <= small:  # circular
        E = nu
        M = nu
    elif small < ecc and ecc <= 1 - small:  # elliptical
        sine = (np.sqrt(1.0 - ecc * ecc) * np.sin(nu)) / \
            (1.0 + ecc * np.cos(nu))
        cose = (ecc + np.cos(nu)) / (1.0 + ecc * np.cos(nu))

        E = np.arctan2(sine, cose)
        M = E - ecc * np.sin(E)

        E = attitude.normalize(E, 0, 2 * np.pi)
        M = attitude.normalize(M, 0, 2 * np.pi)

    elif np.absolute(ecc - 1) <= small:  # parabolic
        B = np.tan(nu / 2)

        E = B
        M = B + 1.0 / 3 * B**3

        # E = revcheck(E);
        # M = revcheck(M);
    elif ecc > 1 + small:  # hyperbolic
        sine = (np.sqrt(ecc**2 - 1) * np.sin(nu)) / \
            (1.0 + ecc * np.cos(nu))
        H = np.arcsinh(sine)
        E = H
        M = ecc * np.sinh(H) - H

        # E = revcheck(E);
        # M = revcheck(M);
    else:
        print("Eccentricity is out of bounds 0 < ecc < inf")

    return (E, M)


def tof_delta_t(p, ecc, mu, nu_0, delta_t):
    """
        Propogate a COE into the future
    """

    tol = 1e-9
    # calculate initial eccentric anomaly and mean anomaly
    E_0, M_0 = nu2anom(nu_0, ecc)

    # check eccentricity
    if np.absolute(ecc - 1) < tol:  # parabolic
        n = 2 * np.sqrt(mu / p**3)
    else:
        a = p / (1 - ecc**2)
        # calculate mean motion
        n = np.sqrt(mu / a**3)

    # calculate mean anomaly after delta t

    M_f = M_0 + n * delta_t
    k = np.floor(M_f / (2 * np.pi))
    M_f = M_f - 2 * np.pi * k
    # calculate eccentric anomaly from mean anomaly (newton iteration)

    E_f, nu_f, count = kepler_eq_E(M_f, ecc)

    return (E_f, M_f, nu_f)


def fpa_solve(nu, ecc):
    """Calculate flight path angle

    Inputs:
        - nu - true anomaly of orbit -2*pi < nu < 2*pi
        - ecc - eccentricity of orbit 0 < ecc < inf

    Outputs:
        - fpa - flight path angle in rad -pi < fpa < pi

    Dependencies:
        - none

    Author:
        - Shankar Kulumani 15 Sept 2012
            - created for AAE532 PS4
        - Shankar Kulumani 29 Sept 2012
            - modified for diferennt orbit types

    References
        - AAE532 Notes
        - Vallado 3rd Ed pg 113
    """
    tol = 1e-9

    if ecc - 1 > tol:  # hyperbolic
        # convert nu to eccentric anomaly
        H, _ = nu2anom(nu, ecc)
        sin_fpa = ecc * np.sinh(H) / np.sqrt(ecc**2 * np.cosh(H)**2 - 1)
        cos_fpa = np.sqrt((ecc**2 - 1) / (ecc**2 * np.cosh(H)**2 - 1))
        fpa = np.arctan2(sin_fpa, cos_fpa)  # rad
    else:
        if np.absolute(ecc - 1) < tol:  # parabolic
            fpa = nu / 2
        else:
            if ecc > tol:  # elliptical
                sin_fpa = ecc * np.sin(nu) / np.sqrt(1 +
                                                     2 * ecc * np.cos(nu) + ecc**2)
                cos_fpa = (1 + ecc * np.cos(nu)) / \
                    np.sqrt(1 + 2 * ecc * np.cos(nu) + ecc**2)

                # atan2 allows for quadrant check
                fpa = np.arctan2(sin_fpa, cos_fpa)  # rad
            else:  # circular
                fpa = 0

    return fpa


def elp_orbit_el(p, ecc, inc, raan, arg_p, nu, mu):
    """Elliptical Orbit Characteristics/Elements

    Purpose:
        - Calculates elliptical orbital parameters using conic equations

    Inputs:
        - p - semi-major axis (km)
        - ecc - eccentricity
        - raan - right acsension of the ascending node (rad) 0 < raan <
        2*pi
        - inc - inclination (rad) 0 < inc < pi
        - arg_p - argument of periapsis (rad) 0 < arg_p < 2*pi
        - nu - true anomaly (rad) 0 < nu < 2*pi
        - mu - gravitational parameter of central body (km^3/sec^2).
        - arglat - argument of latitude(CI) rad 0 <arglat < 2*pi
        - truelon - true longitude (CE) rad 0 < truelon < 2*pi
        - lonper - longitude of periapsis rad 0 < lonper < 2*pi

    Outputs:
        - a - semi-major axis in km
        - h - magnitude of angular momentum vector in km^2/sec
        - period - period of orbit in seconds
        - sme - specific mechanical energy of orbit in km^2/sec^2
        - r_per - radius of periapsis in km
        - r_apo - radius of apoapsis in km
        - r - current radius in km
        - v - current velocity in km/sec
        - v_circ - equivalent circular orbit velocity at current radius in
        km/sec
        - v_esc - escape speed at current radius km/sec
        - fpa - flight path angle in rad 0 < fpa < pi/2
        - E - eccentric anomaly in rad 0 < E < 2*pi
        - M - mean anomaly in rad 0 < M < 2*pi
        - n - mean motion in 1/sec

    Dependencies:
        - fpa_solve - find flight path angle
        - ecc_anomaly - find eccentric and mean anomaly given true
        anomaly
        - aae532_constants - AAE532 class constants
        - ROT3 - simple rotation about third axis
        - use orbit_el as driver function to print results to screen

    Author:
        - Shankar Kulumani 16 Sept 2012
            - used code from AAE532 PS4
        - Shankar Kulumani 19 Sept 2012
            - added escape speed
            - added print capability
            - added constants
            - added mean motion
            - added lvlh and pqw vectors
        - Shankar Kulumani 28 June 2017
            - Convert to Python for MAE3145 class

    References
        - AAE532
        - Any astrodynamics book/internet
    """
    # calculate semi-major axis
    a = p / (1 - ecc**2)  # km
    # angular momentum scalar
    h = np.sqrt(p * mu)  # km^2/sec
    # period of orbit
    period = 2 * np.pi * np.sqrt(a**3 / mu)  # sec
    # specific mechanical energy
    sme = -mu / (2 * a)  # km^2/sec^2

    fpa = fpa_solve(nu, ecc)  # radians

    # radius of periapsis and apoapsis
    r_per = a * (1 - ecc)  # km
    r_apo = a * (1 + ecc)  # km

    # convert to pqw (perifocal frame)
    r_ijk, v_ijk, r_pqw, v_pqw = coe2rv(p, ecc, inc, raan, arg_p, nu, mu)

    r = np.linalg.norm(r_pqw)
    v = np.linalg.norm(v_pqw)

    # convert position and velocity to lvlh frame
    r_lvlh = np.array([r, 0, 0])  # [r_hat theta_hat h_hat] km
    # [r_hat theta_hat h_hat] km/sec
    v_lvlh = v * np.array([np.sin(fpa), np.cos(fpa), 0])

    v_circ = np.sqrt(mu / r)

    v_esc = np.sqrt(2) * v_circ

    E, M = nu2anom(nu, ecc)  # rad
    E = float(E)
    M = float(M)

    n = np.sqrt(mu / a**3)  # mean motion in 1/sec

    return (a, h, period, sme, fpa, r_per, r_apo, r_ijk, v_ijk, r_pqw, v_pqw,
            r_lvlh, v_lvlh, r, v, v_circ, v_esc, E, M, n)


def par_orbit_el(p, ecc, inc, raan, arg_p, nu, mu):

    # calculate semi-latus rectum (km)
    a = np.inf
    # velocity at infinite radius (km/sec)
    v_inf = 0

    # mechanical energy (km^2/sec^2)
    sme = mu / 2 / a

    # angular momentum scalar
    h = np.sqrt(p * mu)  # km^2/sec

    fpa = fpa_solve(nu, ecc)  # radians

    # radius of periapsis and apoapsis
    r_per = p / 2  # km

    r_ijk, v_ijk, r_pqw, v_pqw = coe2rv(p, ecc, inc, raan, arg_p, nu, mu)

    # position and velocity within orbit
    r = norm(r_pqw)  # km
    v = norm(v_pqw)  # km/sec

    # convert position and velocity to lvlh frame
    r_lvlh = np.array([r, 0, 0])  # [r_hat theta_hat h_hat] km
    # [r_hat theta_hat h_hat] km/sec
    v_lvlh = v * np.array([np.sin(fpa), np.cos(fpa), 0])

    v_circ = np.sqrt(mu / r)

    v_esc = np.sqrt(2) * v_circ

    B, M_B = nu2anom(nu, ecc)  # rad

    n = 2 * np.sqrt(mu / p**3)  # mean motion in 1/sec

    return (a, v_inf, sme, h, fpa, r_per, r_ijk, v_ijk,
            r_pqw, v_pqw, r_lvlh, v_lvlh, r, v, v_circ, v_esc, B, M_B, n)


def hyp_orbit_el(p, ecc, inc, raan, arg_p, nu, mu):
    """Hyperbolic Orbit Characteristics/Elements

    Purpose: 
        - calculates orbital parameters for a hyperbolic orbit

    [a, v_inf, b, sme, flyby, nu_inf, h, fpa, r_per, r_ijk, v_ijk, 
    r_pqw, v_pqw, r_lvlh, v_lvlh, r, v, v_circ, v_esc, H, M_H, n] = hyp_orbit_el(
    p, ecc, inc, raan, arg_p, nu, mu)

    Inputs: 
        - 

    Outputs: 
        - List/describe outputs of function

    Dependencies: 
        - coe2rv.m - convert COE to position and velocity vector

    Author: 
        - Shankar Kulumani 29 Sept 2012
            - modified outputs
        - Shankar Kulumani 5 September 2017
            - convert to Python for MAE3145

    References
        - AAE532 and any astrodynamics book 
    """
    # calculate semi-latus rectum (km)
    a = p / (ecc**2 - 1)
    # velocity at infinite radius (km/sec)
    v_inf = np.sqrt(mu / a)
    # aiming radius (semi-minor axis b) (km)
    b = a * np.sqrt(ecc**2 - 1)
    # mechanical energy (km^2/sec^2)
    sme = mu / 2 / a
    # flyby angle (rad)
    flyby = np.arcsin(1 / ecc) * 2
    # true anomaly at infinite radius (rad)
    nu_inf = pi / 2 + flyby / 2

    # angular momentum scalar
    h = np.sqrt(p * mu)  # km^2/sec

    fpa = fpa_solve(nu, ecc)  # radians

    # radius of periapsis and apoapsis
    r_per = a * (ecc - 1)  # km

    r_ijk, v_ijk, r_pqw, v_pqw = coe2rv(p, ecc, inc, raan, arg_p, nu, mu)

    # position and velocity within orbit
    r = np.linalg.norm(r_pqw)  # km
    v = np.linalg.norm(v_pqw)  # km/sec

    # convert position and velocity to lvlh frame
    r_lvlh = np.array([r, 0, 0])  # [r_hat theta_hat h_hat] km
    # [r_hat theta_hat h_hat] km/sec
    v_lvlh = v * np.array([np.sin(fpa), np.cos(fpa), 0])

    v_circ = np.sqrt(mu / r)

    v_esc = np.sqrt(2) * v_circ

    H, M_H = nu2anom(nu, ecc)  # rad

    n = np.sqrt(mu / a**3)  # mean motion in 1/sec

    return (a, v_inf, b, sme, flyby, nu_inf, h, fpa, r_per, r_ijk, v_ijk,
            r_pqw, v_pqw, r_lvlh, v_lvlh, r, v, v_circ, v_esc, H, M_H, n)


def orbit_el(p, ecc, inc, raan, arg_p, nu, mu, print_flag=False):
    """Orbit Characteristics/Elements

    Purpose:
        - Calculates  orbital parameters using conic equations

        orbit_el(p,ecc,inc,raan,arg_p,nu,mu,print_flag)

    Inputs:
        - a - semi-major axis in km
        - ecc - eccentricity of orbit
        - mu - gravitational parameter in km^3/sec^2
        - nu - true anomaly in rad 0 < nu < 2*pi
        - print_flag - 'true' or 'false' to print outputs to screen

    Outputs:
        - none - prints data to screen

    Dependencies:
        - elp_orbit_el.m - elliptical orbit elements
        - hyp_orbit_el.m - hypberbolic orbit elements
        - par_orbit_el.m - parabolic orbit elements

    Author:
        - Shankar Kulumani 16 Sept 2012
            - used code from AAE532 PS4
        - Shankar Kulumani 19 Sept 2012
            - added escape speed
            - added print capability
            - added constants
            - added mean motion
            - added lvlh and pqw vectors
        - Shankar Kulumani 27 Sept 2012
            - modifying to add hyperbolic orbit capability
            - moved elliptical stuff to elp_orbit_el.m
            - removed outputs
        - Shankar Kulumani 29 Sept 2012
            - modified fprintf commands to make it more like STK
        - Shankar Kulumani 5 September 2017
            - modify for Python for MAE3145

    References
        - AAE532 Notes

    """
    # load constants

    r_earth = constants.earth.radius
    km2au = constants.km2au
    rad2deg = constants.rad2deg
    tol = 1e-9

    if ecc < 1:  # elliptical
        (a, h, period, sme, fpa, r_per, r_apo, r_ijk, v_ijk,
         r_pqw, v_pqw, r_lvlh, v_lvlh, r, v, v_circ, v_esc,
         E, M, n) = elp_orbit_el(p, ecc, inc, raan, arg_p, nu, mu)
        # build string for output
        string = '\n'

        string += 'Satellite State \n'
        string += 'Position and Velocity in LVLH frame \n'
        string += '{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n'.format(
            'r_hat:', r_lvlh[0], 'rd_hat:', v_lvlh[0], 't_hat:', r_lvlh[1], 'td_hat:', v_lvlh[1], 'h_hat:', r_lvlh[2], 'hd_hat:', v_lvlh[2])

        string += '\n'
        string += 'Position and Velocity in EPH/PQW frame \n'
        string += '{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n'.format(
            'e_hat:', r_pqw[0], 'ed_hat:', v_pqw[0], 'p_hat:', r_pqw[1], 'pd_hat:', v_pqw[1], 'h_hat:', r_pqw[2], 'hd_hat:', v_pqw[2])

        string += '\n'
        string += 'Position and Velocity in IJK frame \n'
        string += '{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n'.format(
            'i_hat:', r_ijk[0], 'id_hat:', v_ijk[0], 'j_hat:', r_ijk[1], 'jd_hat:', v_ijk[1], 'k_hat:', r_ijk[2], 'kd_hat:', v_ijk[2])

        string += '\n'
        string += '{:10s}: {:20.15g} km = {:20.15g} AU \n'.format(
            'RAD_MAG', r, r * km2au)
        string += '{:10s}: {:20.15g} km/sec \n'.format('VEL_MAG', v)

        string += '\n'
        string += 'Orbital Elements \n'
        string += '{:10s} {:20.15g} km {:10s} {:20.15g} deg\n{:10s} {:20.15g}    {:10s} {:20.15g} deg\n{:10s} {:20.15g} deg {:9s} {:20.15g} deg\n'.format(
            'sma:', a, 'raan:', raan * rad2deg, 'ecc:', ecc, 'arg_p:', arg_p * rad2deg, 'inc:', inc * rad2deg, 'nu:', nu * rad2deg)

        string += '\n'
        string += 'Elliptic Orbital Parameters\n'
        string += '{:10s}: {:20.15g} km = {:20.15g} AU \n'.format(
            'P', p, p * km2au)
        string += '{:10s}: {:20.15g} km^2/sec \n'.format('ANG MOM', h)
        string += '{:10s}: {:20.15g} sec = {:20.15g} hr\n'.format(
            'PERIOD', period, period / 3600)
        string += '{:10s}: {:20.15g} km^2/sec^2 \n'.format('ENGERGY', sme)
        string += '{:10s}: {:20.15g} km = {:20.15g} AU \n'.format(
            'RAD_PER', r_per, r_per * km2au)
        string += '{:10s}: {:20.15g} km = {:20.15g} AU \n'.format(
            'RAD_APO', r_apo, r_apo * km2au)

        string += '\n'
        string += '{:10s}: {:20.15g} km/sec \n'.format('VEL_CIRC', v_circ)
        string += '{:10s}: {:20.15g} km/sec \n'.format('VEL_ESC', v_esc)
        string += '{:10s}: {:20.15g} deg \n'.format('TRUE_ANOM', nu * rad2deg)
        string += '{:10s}: {:20.15g} deg \n'.format('FPA', fpa * rad2deg)
        string += '{:10s}: {:20.15g} deg \n'.format('ECC_ANOM', E * rad2deg)
        string += '{:10s}: {:20.15g} deg \n'.format('MEAN_ANOM', M * rad2deg)
        string += '{:10s}: {:20.15g} deg/sec \n'.format(
            'MEAN_MOT', n * rad2deg)

        string += '\n'
        string += '{:10s}: {:20.15g} sec = {:20.15g} hr \n'.format(
            'T_PAST_PER', 1 / n * M, 1 / n * M / 3600)
    elif (ecc - 1) > tol:  # hyperbolic
        (a, v_inf, b, sme, flyby, nu_inf, h, fpa, r_per, r_ijk, v_ijk, r_pqw,
         v_pqw, r_lvlh, r, v, v_circ, v_esc, H, M_H, n) = hyp_orbit_el(p, ecc, inc, raan, arg_p, nu, mu)
        # build string for output
        string = '\n'

        string += 'Satellite State \n'
        string += 'Position and Velocity in LVLH frame \n'
        string += '{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n'.format(
            'r_hat:', r_lvlh[0], 'rd_hat:', v_lvlh[0], 't_hat:', r_lvlh[1], 'td_hat:', v_lvlh[1], 'h_hat:', r_lvlh[2], 'hd_hat:', v_lvlh[2])

        string += '\n'
        string += 'Position and Velocity in EPH/PQW frame \n'
        string += '{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n'.format(
            'e_hat:', r_pqw[0], 'ed_hat:', v_pqw[0], 'p_hat:', r_pqw[1], 'pd_hat:', v_pqw[1], 'h_hat:', r_pqw[2], 'hd_hat:', v_pqw[2])

        string += '\n'
        string += 'Position and Velocity in IJK frame \n'
        string += '{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n'.format(
            'i_hat:', r_ijk[0], 'id_hat:', v_ijk[0], 'j_hat:', r_ijk[1], 'jd_hat:', v_ijk[1], 'k_hat:', r_ijk[2], 'kd_hat:', v_ijk[2])

        string += '\n'
        string += '{:10s}: {:20.15g} km = {:20.15g} AU \n'.format(
            'RAD_MAG', r, r * km2au)
        string += '{:10s}: {:20.15g} km/sec \n'.format('VEL_MAG', v)

        string += '\n'
        string += 'Orbital Elements \n'
        string += '{:10s} {:20.15g} km {:10s} {:20.15g} deg\n{:10s} {:20.15g}    {:10s} {:20.15g} deg\n{:10s} {:20.15g} deg {:9s} {:20.15g} deg\n'.format(
            'sma:', a, 'raan:', raan * rad2deg, 'ecc:', ecc, 'arg_p:', arg_p * rad2deg, 'inc:', inc * rad2deg, 'nu:', nu * rad2deg)

        string += '\n'
        string += 'Hyperbolic Orbital Parameters\n'
        string += '{:10s}: {:20.15g} km = {:20.15g} AU \n'.format(
            'P', p, p * km2au)
        string += '{:10s}: {:20.15g} km^2/sec \n'.format('ANG MOM', h)
        string += '{:10s}: {:20.15g} km^2/sec^2 \n'.format('ENGERGY', sme)
        string += '{:10s}: {:20.15g} km = {:20.15g} AU \n'.format(
            'RAD_PER', r_per, r_per * km2au)

        string += '\n'
        string += '{:10s}: {:20.15g} km/sec \n'.format('V_INF', v_inf)
        string += '{:10s}: {:20.15g} km \n'.format('RAD_AIM', b)
        string += '{:10s}: {:20.15g} deg \n'.format('FLYBY', flyby * rad2deg)
        string += '{:10s}: {:20.15g} deg \n'.format('NU_INF', nu_inf * rad2deg)

        string += '\n'
        string += '{:10s}: {:20.15g} km/sec \n'.format('VEL_CIRC', v_circ)
        string += '{:10s}: {:20.15g} km/sec \n'.format('VEL_ESC', v_esc)
        string += '{:10s}: {:20.15g} deg \n'.format('TRUE_ANOM', nu * rad2deg)
        string += '{:10s}: {:20.15g} deg \n'.format('FPA', fpa * rad2deg)
        string += '{:10s}: {:20.15g} deg \n'.format('HYP_ANOM', H * rad2deg)
        string += '{:10s}: {:20.15g} deg \n'.format('MEAN_ANOM', M_H * rad2deg)
        string += '{:10s}: {:20.15g} deg/sec \n'.format(
            'MEAN_MOT', n * rad2deg)

        string += '\n'
        string += '{:10s}: {:20.15g} sec = {:20.15g} hr \n'.format(
            'T_PAST_PER', 1 / n * M, 1 / n * M / 3600)
    elif np.absolute(ecc - 1) < tol:  # parabolic
        (a, v_inf, sme, h, fpa, r_per, r_ijk, v_ijk, r_pqw, v_pqw, r_lvlh,
         v_lvlh, r, v, v_circ, v_esc, B, M_B, n) = par_orbit_el(p, ecc, inc,
                                                                raan, arg_p,
                                                                nu, mu)
        # build string for output
        string = '\n'

        string += 'Satellite State \n'
        string += 'Position and Velocity in LVLH frame \n'
        string += '{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n'.format(
            'r_hat:', r_lvlh[0], 'rd_hat:', v_lvlh[0], 't_hat:', r_lvlh[1], 'td_hat:', v_lvlh[1], 'h_hat:', r_lvlh[2], 'hd_hat:', v_lvlh[2])

        string += '\n'
        string += 'Position and Velocity in EPH/PQW frame \n'
        string += '{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n'.format(
            'e_hat:', r_pqw[0], 'ed_hat:', v_pqw[0], 'p_hat:', r_pqw[1], 'pd_hat:', v_pqw[1], 'h_hat:', r_pqw[2], 'hd_hat:', v_pqw[2])

        string += '\n'
        string += 'Position and Velocity in IJK frame \n'
        string += '{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n{:10s} {:20.15g} km {:10s} {:20.15g} km/sec\n'.format(
            'i_hat:', r_ijk[0], 'id_hat:', v_ijk[0], 'j_hat:', r_ijk[1], 'jd_hat:', v_ijk[1], 'k_hat:', r_ijk[2], 'kd_hat:', v_ijk[2])

        string += '\n'
        string += '{:10s}: {:20.15g} km = {:20.15g} AU \n'.format(
            'RAD_MAG', r, r * km2au)
        string += '{:10s}: {:20.15g} km/sec \n'.format('VEL_MAG', v)

        string += '\n'
        string += 'Orbital Elements \n'
        string += '{:10s} {:20.15g} km {:10s} {:20.15g} deg\n{:10s} {:20.15g}    {:10s} {:20.15g} deg\n{:10s} {:20.15g} deg {:9s} {:20.15g} deg\n'.format(
            'sma:', a, 'raan:', raan * rad2deg, 'ecc:', ecc, 'arg_p:', arg_p * rad2deg, 'inc:', inc * rad2deg, 'nu:', nu * rad2deg)

        string += '\n'
        string += 'Parabolic Orbital Parameters\n'
        string += '{:10s}: {:20.15g} km = {:20.15g} AU \n'.format(
            'P', p, p * km2au)
        string += '{:10s}: {:20.15g} km^2/sec \n'.format('ANG MOM', h)
        string += '{:10s}: {:20.15g} km^2/sec^2 \n'.format('ENGERGY', sme)
        string += '{:10s}: {:20.15g} km = {:20.15g} AU \n'.format(
            'RAD_PER', r_per, r_per * km2au)

        string += '\n'
        string += '{:10s}: {:20.15g} km/sec \n'.format('V_INF', v_inf)

        string += '\n'
        string += '{:10s}: {:20.15g} km/sec \n'.format('VEL_CIRC', v_circ)
        string += '{:10s}: {:20.15g} km/sec \n'.format('VEL_ESC', v_esc)
        string += '{:10s}: {:20.15g} deg \n'.format('TRUE_ANOM', nu * rad2deg)
        string += '{:10s}: {:20.15g} deg \n'.format('FPA', fpa * rad2deg)
        string += '{:10s}: {:20.15g} deg \n'.format('HYP_ANOM', B * rad2deg)
        string += '{:10s}: {:20.15g} deg \n'.format('MEAN_ANOM', n * rad2deg)
        string += '{:10s}: {:20.15g} deg/sec \n'.format(
            'MEAN_MOT', n * rad2deg)

        string += '\n'
        string += '{:10s}: {:20.15g} sec = {:20.15g} hr \n'.format(
            'T_PAST_PER', 1 / n * M_B, 1 / n * M_B / 3600)

    if print_flag:
        print(string)

    return string



"""Start of propogator module

Hold many useful functions for numerical integration of astrodynamics
"""
from astro import constants
import numpy as np

def accel_third(mu_j, r_i2j, r_q2j):
    """Third body perturbations

    Calculate relative acceleration of third body (j) on object of
    interest (i) and it's reference object (q)

    Inputs: 
    - mu_j : gravitational parameters of third body in km^3/sec^2
    - r_i2j : position vector from particle of interest to the third body in km
    - r_q2j : position vector from the central body to the third body in km

    Outpus: 
    - direct : direct acceleration in km/sec^2
    - indirect : indirect acceleration in km/sec^2
    - perturbing : sum of direct and indirection

    Dependencies: 
    - None

    Author: 
        - Shankar Kulumani 8 Sept 2012
            - list revisions
        - Shankar Kulumani 25 Sept 2017
            - into Python

    References
        - AAE532 Lesson 6 31 Aug 2012 
        - MAE3145 HW2
    """

    direct = mu_j*r_i2j/np.linalg.norm(r_i2j)**3
    indirect = -mu_j*r_q2j/np.linalg.norm(r_q2j)**3

    perturbing = direct + indirect
    return direct, indirect, perturbing

def accel_twobody(mass_int, mass_body, r_body2int, G=constants.G):
    """Two body dominant acceleration

    """

    accel = -G * ( mass_body + mass_int) / np.linalg.norm(r_body2int)**3 * r_body2int
    return accel



"""Download and handle TLEs
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
# import ephem
from collections import namedtuple

from . import time, kepler, geodetic
from kinematics import attitude

import matplotlib.pyplot as plt
import matplotlib.lines as mlines
import matplotlib.patches as mpatches
import pdb

deg2rad = np.pi / 180
rad2deg = 180 / np.pi
sec2day = 1 / (24 * 3600)
day2sec = 24 * 3600

# tuple to hold all the items from a single TLE
TLE = namedtuple('TLE', [
    'satname', 'satnum', 'classification', 'id_year', 'id_launch',
    'id_piece', 'epoch_year', 'epoch_day', 'ndot_over_2', 'nddot_over_6',
    'bstar', 'ephtype', 'elnum', 'checksum1',
    'inc', 'raan', 'ecc', 'argp', 'ma', 'mean_motion', 'epoch_rev',
    'checksum2', 'good'])

COE = namedtuple('COE', ['n', 'ecc', 'raan', 'argp', 'mean', 'E', 'nu', 'a',
                         'p', 'inc'])

PASS = namedtuple('PASS', ['jd', 'az', 'el', 'site_eci', 'sat_eci', 'gst', 'lst',
                           'sun_alt', 'alpha', 'beta', 'sun_eci', 'rho'])

RADAR_PASS = namedtuple('RADAR_PASS', ['jd', 'site_eci', 'sat_r_eci', 'sat_v_eci', 
                                       'gst', 'lst', 'sun_eci', 'sun_alt', 'alpha', 
                                       'beta', 'rho', 'az', 'el', 'drho', 'daz', 'dele'])

def get_tle_spacetrack(filename, flag='all'):
    r"""Download TLEs from spacetrack.org

    This function will download some TLEs from Space - track.org. It requires a
    valid user account on the website. It uses favorites lists which can be
    setup in your own account.

    Parameters
    ----------
    filename : string
        Filename to store the downloaded TLEs
    flag : string
        This string can take one of three values. It allows you decide which
        TLE set to download from the website.
        all - Download bulk catalog
        visible - Download the visible satellite list
        testing - Download personal list called Testing
        rv2coe - list used in RV2COE
        comfix - list used in COMFIX

    Returns
    -------
    None

    Other Parameters
    ----------------
    None

    Raises
    ------
    None

    See Also
    --------
    None

    Notes
    -----

    References
    ----------

    .. [1] spacetrack - https://github.com/python-astrodynamics/spacetrack

    Examples
    --------

    """
    from spacetrack import SpaceTrackClient

    password = input('Enter SpaceTrack.org password: ')
    st = SpaceTrackClient('shanks.k', password)
    with open(filename, 'w') as f:
        if flag == 'all':
            all_tles = st.tle_latest(ordinal=1, format='3le')
        elif flag == 'testing':
            all_tles = st.tle_latest(
                favorites='Testing', ordinal=1, format='3le')
        elif flag == 'visible':
            all_tles = st.tle_latest(
                favorites='Visible', ordinal=1, format='3le')
        elif flag == 'rv2coe':
            all_tles = st.tle_latest(favorites='RV2COE', ordinal=1,
                                     format='3le')
        elif flag == 'comfix':
            all_tles = st.tle_latest(favorites='COMFIX', ordinal=1,
                                     format='3le')
        else:
            print("Incorrect TLE favorites flag")
            all_tles = "Incorrect flag"

        f.write(all_tles)


# def get_tle_ephem(filename):
#     """Load TLEs from a file
#     """
#     satlist = []
#     with open(filename, 'r') as f:
#         l1 = f.readline()
#         while l1:
#             l2 = f.readline()
#             l3 = f.readline()
#             sat = ephem.readtle(l1, l2, l3)
#             satlist.append(sat)
#             print(sat.name)
#             l1 = f.readline()

#     print("{} satellites loaded into list".format(len(satlist)))
#     return satlist

# write my own TLE parser since ephem doesn't save all the values


def validtle(l0, l1, l2):
    r"""Ensure 3 line TLEs are valid.

    This function will check each line of a 3 line TLE and ensure they are
    valid. It assumes the format is that which is found on space-track.org.

    Parameters
    ----------
    l0 : string
        First line of TLE. This line holds the object name.
        Sometimes is starts with a 0 and other times it does not.
        Both will be valid in this code.
    l1: string
        Second line of TLE. This is first real line with real data.
        Should begin with a 1 and pass a checksum. There is also a line length
        requirement.
    l2: string
        Third line of TLE. This also contains data. Begins with a 2 and has a
        checksum and line lenght requirement.

    Returns
    -------
    boolean:
        This returns a boolean object which is the concatenation of all the
        requirements for the checking.  If this output is True then the three
        lines of the inputted TLE are probably valid.

    Other Parameters
    ----------------
        None

    Raises
    ------
        None

    See Also
    --------
    checksum: another related function to calculate the line checksum
    parsetle: this will take the three lines and extract out the TLE data
    stringScientificNotationToFloat: Converts the TLE number format to float

    Notes
    -----
    TLEs are an old format, but they are well documented. In addition, the TLE
    has some inherent limitations on the expected accuracy of the ephemerides
    and the data contained within it.

    References
    ----------

    .. [1] VALLADO, David A. Fundamentals of Astrodynamics and Applications. 3
    ed. Microcosm Press, 2007.
    .. [2] http://celestrak.com/NORAD/documentation/tle-fmt.asp

    Examples
    --------

    """
    l1_valid_start = int(l1[0]) == 1
    l2_valid_start = int(l2[0]) == 2

    l1_valid_length = len(l1) == 69
    l2_valid_length = len(l2) == 69

    l1_valid_checksum = int(l1[-1]) == checksum(l1)
    l2_valid_checksum = int(l2[-1]) == checksum(l2)
    return (l1_valid_start and l2_valid_start and
            l1_valid_length and l2_valid_length and
            l1_valid_checksum and l1_valid_checksum and l2_valid_checksum)


def checksum(line):
    """The checksums for each line are calculated by adding the all numerical
    digits on that line, including the line number. One is added to the
    checksum for each negative sign (-) on that line. All other non-digit
    characters are ignored.  @note this excludes last char for the checksum
    thats already there.
    """
    L = line.strip()
    cksum = 0
    for i in range(68):
        c = L[i]
        if c <= '9' and c >= '0':
            cksum = cksum + int(c)
        elif c == '-':
            cksum = cksum + 1
        else:
            continue

    cksum = cksum % 10

    return cksum

def stringScientificNotationToFloat(sn):
    """Specific format is 5 digits, a + or -, and 1 digit, ex: 01234-5 which is
    0.01234e-5
    """
    return 1e-5 * float(sn[:6]) * 10**int(sn[6:])


def parsetle(l0, l1, l2):
    """
    The function parsetle extracts the elements of a two line element set.
    This takes in the three lines of a TLE and parses the values.

    Inputs:
    l0 - first line consiting of the name of TLE (optional)
    l1 - second line
    l2 - third line

    Outputs - namedtuple TLE:
    satname - satellite name
    satnum - satellite number found (0 if the requested sat was not found)
    classfication  - classification (classified, unclassified)
    id_year - launch year (last 2 digits)
    id_launch - launch number of the year
    id_piece - piece of launch (a string 3 characters long)
    epoch_year - epoch year (last 2 digits)
    epoch_day - epoch day of year
    ndot_over_2 - first time derive of mean mot. divided by 2 (rev/day^2)
    nddot_over_6 - 2nd deriv of mean mot div by 6 (rev/day^3)
    bstar - bstar drag term
    ephtype- ephemeris type
    elnum  - element number (modulo 1000)
    checksum1 - checksum of first line

    inc    - inclination (deg)
    raan   - right ascension of asc. node (deg)
    ecc    - eccentricity
    argp    - argument of perigee (deg)
    ma     - mean anomaly (deg)
    mean_motion - mean motion (revs per day)
    epoch_rev - revolution number(modulo 100,000)
    checksum2 - checksum of second line

    References:
    "Fundamental of Astrodynamics and Applictions, 2nd Ed", by Vallado
    www.celestrak.com/NORAD/documentation/tle-fmt.asp
    manpages.debian.net/cgi-bin/...
    display_man.cgi?id=e2095ebad4eb81b943fcdd55ec9b7521&format=html
    """
    try:
        # parse line zero
        satname = l0[0:23]

        # parse line one
        satnum = int(l1[2:7])
        classification = l1[7:8]
        id_year = int(l1[9:11])
        id_launch = int(l1[11:14])
        id_piece = l1[14:17]
        epoch_year = int(l1[18:20])
        epoch_day = float(l1[20:32])
        ndot_over_2 = float(l1[33:43])
        nddot_over_6 = stringScientificNotationToFloat(l1[44:52])
        bstar = stringScientificNotationToFloat(l1[53:61])
        ephtype = int(l1[62:63])
        elnum = int(l1[64:68])
        checksum1 = int(l1[68:69])

        # parse line 2
        # satellite        = int(line2[2:7])
        inc = float(l2[8:16])
        raan = float(l2[17:25])
        ecc = float(l2[26:33]) * 0.0000001
        argp = float(l2[34:42])
        ma = float(l2[43:51])
        mean_motion = float(l2[52:63])
        epoch_rev = float(l2[63:68])
        checksum2 = float(l2[68:69])

        good = True
    except: # We have a bad TLE but we're not sure where
        # parse line zero
        satname = l0[0:23]

        # parse line one
        satnum = 0
        classification = 'U'
        id_year = 0
        id_launch = 0
        id_piece = 'A'
        epoch_year = 0 
        epoch_day = 0
        ndot_over_2 = 0
        nddot_over_6 = 0
        bstar = 0
        ephtype = 0
        elnum = 0
        checksum1 = 0

        # parse line 2
        # satellite        = int(line2[2:7])
        inc = 0
        raan = 0
        ecc = 0
        argp = 0
        ma = 0
        mean_motion = 0
        epoch_rev = 0
        checksum2 = 0

        good = False

    return TLE(satnum=satnum, classification=classification, id_year=id_year,
               id_launch=id_launch, id_piece=id_piece, epoch_year=epoch_year,
               epoch_day=epoch_day, ndot_over_2=ndot_over_2,
               nddot_over_6=nddot_over_6, bstar=bstar, ephtype=ephtype,
               elnum=elnum, checksum1=checksum1, inc=inc, raan=raan, ecc=ecc,
               argp=argp, ma=ma, mean_motion=mean_motion, epoch_rev=epoch_rev,
               checksum2=checksum2, satname=satname, good=good)


def j2dragpert(inc0, ecc0, n0, ndot2, mu=398600.5, re=6378.137, J2=0.00108263):
    """
    This file calculates the rates of change of the right ascension of the
    ascending node(raandot), argument of periapsis(argdot), and
    eccentricity(eccdot).  The perturbations are limited to J2 and drag only.

    Author:   C2C Shankar Kulumani   USAFA/CS-19   719-333-4741
            12 5 2014 - modified to remove global varaibles
            17 Jun 2017 - Now in Python for awesomeness

    Inputs:
    inc0 - initial inclination (radians)
    ecc0 - initial eccentricity (unitless)
    n0 - initial mean motion (rad/sec)
    ndot2 - mean motion rate divided by 2 (rad/sec^2)

    Outputs:
    raandot - nodal rate (rad/sec)
    argdot - argument of periapsis rate (rad/sec)
    eccdot - eccentricity rate (1/sec)

    Globals: None

    Constants:
    mu - 398600.5 Earth gravitational parameter in km^3/sec^2
    re - 6378.137 Earth radius in km
    J2 - 0.00108263 J2 perturbation factor

    Coupling: None

    References:
    Vallado
    """

    # calculate initial semi major axis and semilatus rectum
    a0 = (mu / n0**2) ** (1 / 3)
    p0 = a0 * (1 - ecc0**2)

    # mean motion with J2
    nvec = n0 * (1 + 1.5 * J2 * (re / p0)**2 *
                 np.sqrt(1 - ecc0**2) * (1 - 1.5 * np.sin(inc0)**2))

    # eccentricity rate
    eccdot = (-2 * (1 - ecc0) * 2 * ndot2) / (3 * nvec)

    # calculate nodal rate
    raandot = (-1.5 * J2 * (re / p0)**2 * np.cos(inc0)) * nvec

    # argument of periapsis rate
    argdot = (1.5 * J2 * (re / p0)**2 * (2 - 2.5 * np.sin(inc0)**2)) * nvec

    return (raandot, argdot, eccdot)


def get_tle(filename):
    """Assuming a file with 3 Line TLEs is given this will parse the file
    and save all the elements to a list or something.
    """
    sats = []
    tles = 0
    lines = 0
    with open(filename, 'r') as f:
        l0 = f.readline().strip()
        while l0:
            l1 = f.readline().strip()
            lines += 1
            l2 = f.readline().strip()
            lines += 1
            # error check to make sure we have read a complete TLE
            if validtle(l0, l1, l2):
                # now we parse the tle
                elements = parsetle(l0, l1, l2)
                tles += 1
                # final check to see if we got a good TLE
                if elements.good:
                    # instantiate a bunch of instances of a Satellite class
                    sats.append(Satellite(elements))

            else:
                print("Invalid TLE")

            l0 = f.readline().strip()
            lines += 1

    print("Read {} lines which should be {} TLEs".format(lines, lines / 3))
    print("Found {} TLEs".format(tles))
    print("Parsed {} SATs".format(len(sats)))
    return sats


class Satellite(object):
    """Satellite class storing all the necessary information for a single
    satellite extracted from a TLE from space-track.org
    """

    def __init__(self, elements):
        # now do some conversions of the TLE components
        epoch_year = (2000 + elements.epoch_year
                      if elements.epoch_year < 70
                      else 1900 + elements.epoch_year)
        # working units
        ndot2 = elements.ndot_over_2 * (2 * np.pi) * (sec2day**2)
        inc0 = elements.inc * deg2rad
        raan0 = elements.raan * deg2rad
        argp0 = elements.argp * deg2rad
        mean0 = elements.ma * deg2rad
        n0 = elements.mean_motion * 2 * np.pi * sec2day
        ecc0 = elements.ecc
        epoch_day = elements.epoch_day

        # calculate perturbations raan, argp, ecc
        raandot, argpdot, eccdot = j2dragpert(inc0, ecc0, n0, ndot2)

        # calculate epoch julian date
        mo, day, hour, mn, sec = time.dayofyr2mdhms(epoch_year, epoch_day)
        epoch_jd, _ = time.date2jd(epoch_year, mo, day, hour, mn, sec)

        # store with class
        self.satname = elements.satname
        self.satnum = elements.satnum
        self.tle = elements
        self.ndot2 = ndot2
        self.inc0 = inc0
        self.raan0 = raan0
        self.ecc0 = ecc0
        self.argp0 = argp0
        self.mean0 = mean0
        self.n0 = n0
        self.raandot = raandot
        self.argpdot = argpdot
        self.eccdot = eccdot
        self.epoch_jd = epoch_jd
        self.epoch_year = epoch_year
        self.epoch_day = epoch_day

    def tle_update(self, jd_span, mu=398600.5):
        """Update the state of satellite given a JD time span

        This procedure uses the method of general perturbations to update
        classical elements from epoch to a future time for inclined elliptical
        orbits.  It includes the effects due to first order secular rates
        (second order for mean motion) caused by drag and J2.

        Author:   C2C Shankar Kulumani   USAFA/CS-19   719-333-4741

        Inputs:
            deltat - elapsed time since epoch (sec)
            n0 - mean motion at epoch (rad/sec)
            ndot2 - mean motion rate divided by 2 (rad/sec^2)
            ecc0 - eccentricity at epoch (unitless)
            eccdot - eccentricity rate (1/sec)
            raan0 - RAAN at epoch (rad)
            raandot - RAAN rate (rad/sec)
            argp0 - argument of periapsis at epoch (rad)
            argdot - argument of periapsis rate (rad/sec)
            mean0 - mean anomaly at epoch

        Outputs:
            n - mean motion at epoch + deltat (rad/sec)
            ecc -  eccentricity at epoch + deltat (unitless)
            raan - RAAN at epoch + deltat (rad)
            argp - argument of periapsis at epoch + deltat (rad)
            nu -  true anomaly at epoch + deltat (rad)

        Globals: None

        Constants: None

        Coupling:
            newton.m

        References:
            Astro 321 PREDICT LSN 24-25
        """

        deltat = (jd_span - self.epoch_jd) * day2sec
        n0 = self.n0
        ecc0 = self.ecc0
        ndot2 = self.ndot2
        eccdot = self.eccdot
        raan0 = self.raan0
        raandot = self.raandot
        argp0 = self.argp0
        argpdot = self.argpdot
        mean0 = self.mean0

        # mean motion at future time
        n = n0 + ndot2 * 2 * deltat
        ecc = ecc0 + eccdot * deltat
        raan = raan0 + raandot * deltat
        argp = argp0 + argpdot * deltat

        # get true anomaly at future time
        mean = mean0 + n0 * deltat + ndot2 * deltat**2
        mean = attitude.normalize(mean, 0, 2 * np.pi)

        E, nu, count = kepler.kepler_eq_E(mean, ecc)

        a = (mu / n**2)**(1 / 3)
        p = a * (1 - ecc**2)
        inc = self.inc0 * np.ones_like(p)

        # convert to ECI
        r_eci, v_eci, _, _ = kepler.coe2rv(p, ecc, inc, raan, argp, nu, mu)

        # save all of the variables to the object
        self.jd_span = jd_span
        self.coe = COE(n=n, ecc=ecc, raan=raan, argp=argp, mean=mean,
                       E=E, nu=nu, a=a, p=p, inc=inc)
        self.r_eci = r_eci
        self.v_eci = v_eci
        return 0

    def visible(self, site):
        """Check if current sat is visible from the site
        """
        sat_eci = self.r_eci
        site_eci = site['eci']
        sun_eci = site['sun_eci']

        alpha = np.arccos(np.einsum('ij,ij->i', site_eci, sun_eci) /
                          np.linalg.norm(site_eci, axis=1) / np.linalg.norm(sun_eci,
                                                                            axis=1))
        beta = np.arccos(np.einsum('ij,ij->i', sat_eci, sun_eci) /
                         np.linalg.norm(sat_eci, axis=1) / np.linalg.norm(sun_eci,
                                                                          axis=1))
        sun_alt = np.linalg.norm(sun_eci, axis=1) * np.sin(beta)

        jd_vis = []
        rho_vis = []
        az_vis = []
        el_vis = []

        # create array to store all the passes
        all_passes = []
        current_pass = PASS(jd=[], az=[], el=[], site_eci=[], sat_eci=[],
                            gst=[], lst=[], sun_alt=[], alpha=[], beta=[],
                            sun_eci=[], rho=[])

        jd_last = self.jd_span[0]
        for (jd, a, b, sa, si, su, su_alt, lst, gst) in zip(self.jd_span, alpha,
                                                            beta, sat_eci,
                                                            site_eci, sun_eci,
                                                            sun_alt, site['lst'],
                                                            site['gst']):

            if a > np.pi / 2:
                rho, az, el = geodetic.rhoazel(sa, si, site['lat'], lst)

                if rho < 1500 and el > 10 * deg2rad:
                    if b < np.pi / 2 or su_alt > 6378.137:
                        jd_vis.append(jd)
                        rho_vis.append(rho)
                        az_vis.append(az)
                        el_vis.append(el)

                        # we have a visible state - save to current pass if
                        # time since last visible is small enough if not then
                        # we have to create another pass namedtuple
                        if np.absolute(jd_last - jd) > (10 / 24 / 60):
                            if current_pass.jd:  # new pass
                                all_passes.append(current_pass)
                                current_pass = PASS(jd=[], az=[], el=[],
                                                    site_eci=[], sat_eci=[],
                                                    gst=[], lst=[], sun_alt=[],
                                                    alpha=[], beta=[],
                                                    sun_eci=[], rho=[])

                        current_pass.jd.append(jd)
                        current_pass.az.append(az)
                        current_pass.el.append(el)
                        current_pass.site_eci.append(si)
                        current_pass.sat_eci.append(sa)
                        current_pass.gst.append(gst)
                        current_pass.lst.append(lst)
                        current_pass.sun_alt.append(su_alt)
                        current_pass.alpha.append(a)
                        current_pass.beta.append(b)
                        current_pass.sun_eci.append(su)
                        current_pass.rho.append(rho)
                        jd_last = jd

        all_passes.append(current_pass)

        self.jd_vis = jd_vis
        self.rho_vis = rho_vis
        self.az_vis = az_vis
        self.el_vis = el_vis
        self.pass_vis = all_passes

    def visible_radar(self, site):
        """Check if current sat is visible from the site given an example radar sensor
        """
        sat_r_eci = self.r_eci
        sat_v_eci = self.v_eci
        site_eci = site['eci']
        sun_eci = site['sun_eci']

        alpha = np.arccos(np.einsum('ij,ij->i', site_eci, sun_eci) /
                          np.linalg.norm(site_eci, axis=1) / np.linalg.norm(sun_eci,
                                                                            axis=1))
        beta = np.arccos(np.einsum('ij,ij->i', sat_r_eci, sun_eci) /
                         np.linalg.norm(sat_r_eci, axis=1) / np.linalg.norm(sun_eci,
                                                                          axis=1))
        sun_alt = np.linalg.norm(sun_eci, axis=1) * np.sin(beta)

        jd_vis = []
        rho_vis = []
        az_vis = []
        el_vis = []
        drho_vis = []
        daz_vis = []
        dele_vis = []

        # create array to store all the passes
        all_passes = []
        current_pass = RADAR_PASS(jd=[], site_eci=[], sat_r_eci=[],
                                  sat_v_eci=[], gst=[], lst=[], sun_alt=[],
                                  alpha=[], beta=[], sun_eci=[], 
                                  rho=[], az=[], el=[], drho=[], daz=[], dele=[])

        jd_last = self.jd_span[0]
        for (jd, a, b, sar, sav, si, su, su_alt, lst, gst) in zip(self.jd_span, alpha,
                                                            beta, sat_r_eci, sat_v_eci,
                                                            site_eci, sun_eci,
                                                            sun_alt, site['lst'],
                                                            site['gst']):

            rho, az, el, drho, daz, dele = geodetic.rv2rhoazel(sar, sav, site['lat'],
                                                               site['lon'], site['alt'], jd)
            if el > 10 * deg2rad:
                jd_vis.append(jd)
                rho_vis.append(rho)
                az_vis.append(az)
                el_vis.append(el)
                drho_vis.append(drho)
                daz_vis.append(daz)
                dele_vis.append(dele)

                # we have a visible state - save to current pass if
                # time since last visible is small enough if not then
                # we have to create another pass namedtuple
                if np.absolute(jd_last - jd) > (10 / 24 / 60):
                    if current_pass.jd:  # new pass
                        all_passes.append(current_pass)
                        current_pass = RADAR_PASS(jd=[], site_eci=[], sat_r_eci=[],
                                  sat_v_eci=[], gst=[], lst=[], sun_alt=[],
                                  alpha=[], beta=[], sun_eci=[], 
                                  rho=[], az=[], el=[], drho=[], daz=[], dele=[])

                current_pass.jd.append(jd)
                current_pass.site_eci.append(si)
                current_pass.sat_r_eci.append(sar)
                current_pass.sat_v_eci.append(sav)
                current_pass.gst.append(gst)
                current_pass.lst.append(lst)
                current_pass.sun_alt.append(su_alt)
                current_pass.alpha.append(a)
                current_pass.beta.append(b)
                current_pass.sun_eci.append(su)
                current_pass.rho.append(rho)
                current_pass.az.append(az)
                current_pass.el.append(el)
                current_pass.drho.append(drho)
                current_pass.daz.append(daz)
                current_pass.dele.append(dele)
                jd_last = jd

        all_passes.append(current_pass)

        self.jd_vis = jd_vis
        self.rho_vis = rho_vis
        self.az_vis = az_vis
        self.el_vis = el_vis
        self.drho_vis = drho_vis
        self.daz_vis = daz_vis
        self.dele_vis = dele_vis
        self.pass_vis = all_passes

    def output(self, filename):
        """Write to output file
        """
        space = '    '
        with open(filename, 'w') as f:

            f.write('%s %05.0f\n' % (self.satname, self.satnum))
            f.write(
                'PASS    MON/DAY    HR:MIN(UT)    RHO(KM)    AZ(DEG)    EL(DEG)    SAT          \n')
            f.write(
                '-------------------------------------------------------------------------------\n\n')

            for ii, cur_pass in enumerate(self.pass_vis):
                for jd, rho, az, el in zip(cur_pass.jd, cur_pass.rho, cur_pass.az, cur_pass.el):
                    # convert julian day to normal date
                    yr, mo, day, hr, mn, sec = time.jd2date(jd)

                    # if sec > 30:
                    #     mn = mn + 1
                    #     if mn == 60:
                    #         hr = hr + 1
                    #         mn = 0
                    f.write('%4.0f%s' % (ii, space))
                    f.write('%3.0f/%3.0f%s' % (mo, day, space))
                    f.write('%02.0f:%02.0f:%3.1f%s' % (hr, mn, sec, space))
                    f.write('%7.2f%s' % (rho, space))
                    f.write('%7.2f%s' % (az * 180 / np.pi, space))
                    f.write('%7.2f%s' % (el * 180 / np.pi, space))
                    f.write('%13s\n' % (self.satname))

    def plot_pass(self, pass_num):
        """Try and plot a pass on a polar plot
        """
        def mapr(r):
            return 90 - r

        def add_arrow_to_line2D(
                axes, line, arrow_locs=[0.2, 0.4, 0.6, 0.8],
                arrowstyle='-|>', arrowsize=1, transform=None):
            """
            Add arrows to a matplotlib.lines.Line2D at selected locations.

            Parameters:
            -----------
            axes:
            line: Line2D object as returned by plot command
            arrow_locs: list of locations where to insert arrows, % of total length
            arrowstyle: style of the arrow
            arrowsize: size of the arrow
            transform: a matplotlib transform instance, default to data coordinates

            Returns:
            --------
            arrows: list of arrows
            """
            if not isinstance(line, mlines.Line2D):
                raise ValueError("expected a matplotlib.lines.Line2D object")
            x, y = line.get_xdata(), line.get_ydata()

            arrow_kw = {
                "arrowstyle": arrowstyle,
                "mutation_scale": 10 * arrowsize,
            }

            color = line.get_color()
            use_multicolor_lines = isinstance(color, np.ndarray)
            if use_multicolor_lines:
                raise NotImplementedError("multicolor lines not supported")
            else:
                arrow_kw['color'] = color

            linewidth = line.get_linewidth()
            if isinstance(linewidth, np.ndarray):
                raise NotImplementedError("multiwidth lines not supported")
            else:
                arrow_kw['linewidth'] = linewidth

            if transform is None:
                transform = axes.transData

            arrows = []
            for loc in arrow_locs:
                s = np.cumsum(np.sqrt(np.diff(x) ** 2 + np.diff(y) ** 2))
                n = np.searchsorted(s, s[-1] * loc)
                arrow_tail = (x[n], y[n])
                arrow_head = (np.mean(x[n:n + 2]), np.mean(y[n:n + 2]))
                p = mpatches.FancyArrowPatch(
                    arrow_tail, arrow_head, transform=transform,
                    **arrow_kw)
                axes.add_patch(p)
                arrows.append(p)
            return arrows

        jd = self.pass_vis[pass_num].jd
        az = self.pass_vis[pass_num].az
        el = self.pass_vis[pass_num].el

        sy, smo, sd, sh, smn, ss = time.jd2date(jd[0])
        ey, emo, ed, eh, emn, es = time.jd2date(jd[-1])
        fig = plt.figure()
        ax = plt.subplot(111, projection='polar')
        line = ax.plot(az, mapr(np.rad2deg(el)))[0]
        ax.set_yticks(range(0, 90, 10))
        ax.set_yticklabels(map(str, range(90, 0, -10)))
        ax.set_theta_zero_location("N")
        fig.suptitle("%s" %
                     (self.satname), y=1.05)
        plt.title('%2d/%2d' % (smo, sd))
        plt.title('Start\n%2d:%2d:%3.1f' % (sh, smn, ss), loc='left')
        plt.title('End\n%2d:%2d:%3.1f' % (eh, emn, es), loc='right')
        add_arrow_to_line2D(ax, line, arrow_locs=[0.25, 0.5, 0.75],
                            arrowstyle='->')
        plt.show()



"""This module will download kernels and setup SPICE
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import os
import time
from urllib.request import urlretrieve
import pdb
import spiceypy as spice


cwd = os.path.realpath(os.path.dirname(__file__))
directory = 'kernels'

if not os.path.isdir(os.path.join(cwd, directory)):
    os.mkdir(os.path.join(cwd, directory))


def getKernelNameFromUrl(url):
    """Extract the Kernal name from a URL
    """
    return str(url.split('/')[-1])

def getPathfromUrl(url):
    """Extract the path from the url
    """
    return str(os.path.join(cwd, directory, getKernelNameFromUrl(url)))

def cleanupFile(path):
    """Delete a file from the given path
    """

class NearKernels(object):
    """List of urls and kernels for the Near mission

	This only downloads data for 2001, not the whole mission.

    More data on Near is available:
    https://pdssbn.astro.umd.edu/data_sb/missions/near/index.shtml
    """

    def __init__(self):
        """Construct all the member variables for NEAR
        """
        self.near_id = '-93'
        self.eros_id = '2000433'
        
        self.near_body_frame = 'NEAR_SC_BUS_PRIME'
        self.near_body_frame_id = -93000 
        self.near_msi_frame = 'NEAR_MSI'
        self.near_msi_frame_id = -93001
        self.eros_body_frame = 'IAU_EROS'
        self.eros_body_frame_id = 2000433

        self.inertial_frame = 'J2000'
            
        # mission start from near_171.tsc in UTC
        self.start_et = -122138129.0388301
        self.start_utc = '1996-02-17 20:43:28.775 UTC' 
        
        self.Lsk_url = 'https://naif.jpl.nasa.gov/pub/naif/generic_kernels/lsk/a_old_versions/naif0008.tls'
        self.Ck_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/ck/near_20010101_20010228_v01.bc'
        self.Sclk_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/sclk/near_171.tsc'

        self.PckEros1_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/pck/erosatt_1998329_2001157_v01.bpc'
        self.PckEros2_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/pck/erosatt_1999304_2001151.bpc'
        self.Pck3_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/pck/pck00010.tpc'

        self.Fk_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/fk/eros_fixed.tf'

        self.Ikgrs_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/ik/grs12.ti'
        self.Ikmsi_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/ik/msi15.ti'
        self.Iknis_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/ik/nis14.ti'
        self.Iknlr_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/ik/nlr04.ti'
        self.Ikxrs_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/ik/xrs12.ti'
    
        self.SpkPlanet_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/spk/de403s.bsp'
        self.SpkEros_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/spk/eros80.bsp'
        self.SpkEros2_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/spk/erosephem_1999004_2002181.bsp'
        self.SpkMath_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/spk/math9749.bsp'
        self.SpkNearLanded_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/spk/near_eroslanded_nav_v1.bsp'
        self.SpkNearOrbit_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/spk/near_erosorbit_nav_v1.bsp'
        self.SpkStations_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/spk/stations.bsp'

        # get the path for each kernel
        self.Lsk = getPathfromUrl(self.Lsk_url)
        self.Ck = getPathfromUrl(self.Ck_url)
        self.Sclk = getPathfromUrl(self.Sclk_url)

        self.PckEros1 = getPathfromUrl(self.PckEros1_url)
        self.PckEros2 = getPathfromUrl(self.PckEros2_url) 
        self.Pck3 = getPathfromUrl(self.Pck3_url) 

        self.Fk = getPathfromUrl(self.Fk_url) 

        self.Ikgrs = getPathfromUrl(self.Ikgrs_url)
        self.Ikmsi = getPathfromUrl(self.Ikmsi_url)
        self.Iknis = getPathfromUrl(self.Iknis_url)
        self.Iknlr = getPathfromUrl(self.Iknlr_url)
        self.Ikxrs = getPathfromUrl(self.Ikxrs_url)
    
        self.SpkPlanet = getPathfromUrl(self.SpkPlanet_url)
        self.SpkEros = getPathfromUrl(self.SpkEros_url)
        self.SpkEros2 = getPathfromUrl(self.SpkEros2_url)
        self.SpkMath = getPathfromUrl(self.SpkMath_url)
        self.SpkNearLanded = getPathfromUrl(self.SpkNearLanded_url)
        self.SpkNearOrbit = getPathfromUrl(self.SpkNearOrbit_url)
        self.SpkStations = getPathfromUrl(self.SpkStations_url)
        
        self.urlList = [self.Lsk_url, self.Ck_url, self.Sclk_url, self.Pck3_url, 
                self.PckEros1_url, self.PckEros2_url, self.Fk_url,
                self.Ikgrs_url, self.Ikmsi_url, self.Iknis_url, self.Iknlr_url, 
                self.Ikxrs_url, self.SpkPlanet_url, self.SpkEros_url, 
                self.SpkEros2_url, self.SpkMath_url, self.SpkNearLanded_url, 
                self.SpkNearOrbit_url, self.SpkStations_url]

        self.kernelList = [self.Lsk, self.Ck, self.Sclk, self.Pck3, self.PckEros1, 
                self.PckEros2, self.Fk, self.Ikgrs, self.Ikmsi, self.Iknis, 
                self.Iknlr, self.Ikxrs, self.SpkPlanet, self.SpkEros, 
                self.SpkEros2, self.SpkMath, self.SpkNearLanded, 
                self.SpkNearOrbit, self.SpkStations]

        self.nameList = [getKernelNameFromUrl(url) for url in self.urlList]
        self.kernelDescription = 'Metal Kernel for 2001 NEAR orbit and landing'
        getKernels(self)
        self.metakernel = writeMetaKernel(self, 'near2001.tm')

        self.info()

    def info(self):
        """Read and Output info about the loaded kernels
        """
        spice.kclear()

        spice.furnsh(self.metakernel)

        self.spkList = [self.SpkPlanet, self.SpkEros, self.SpkEros2,
                self.SpkMath, self.SpkNearLanded, self.SpkNearOrbit,
                self.SpkStations]

        self.ckList = [self.Ck]
        self.pckList = [self.PckEros1, self.PckEros2]

        # check SPK coverage
        self.bodies = {}
        self.bodies_coverage = {}
        for spk in self.spkList:
            idcell = spice.spkobj(spk)
            for code in idcell:
                cover = spice.stypes.SPICEDOUBLE_CELL(1000)
                spice.spkcov(spk, code, cover)
                self.bodies[str(code)] = spice.bodc2n(code)
                self.bodies[spice.bodc2n(code)] = code
                self.bodies_coverage[str(code)] = [x for x in cover]
                self.bodies_coverage[spice.bodc2n(code)] = [x for x in cover]

        # check CK coverage
        self.ckframes = {}
        self.ckframes_coverage = {}
        for ck in self.ckList:
            idcell = spice.ckobj(ck)
            for code in idcell:
                cover = spice.ckcov(ck, code, True, 'SEGMENT', 0.0, 'TDB')
                self.ckframes[str(code)] = spice.frmnam(code)
                self.ckframes[spice.frmnam(code)] = code
                self.ckframes_coverage[str(code)] = [x for x in cover]
                self.ckframes_coverage[spice.frmnam(code)] = [x for x in cover]

        # check pck coverage
        self.pckframes = {}
        for pck in self.pckList:
            ids = spice.stypes.SPICEINT_CELL(1000)
            spice.pckfrm(pck, ids)
            for code in ids:
                self.pckframes[str(code)] = spice.frmnam(code)
                self.pckframes[spice.frmnam(code)] = code

        spice.kclear()
        

class CassiniKernels(object):
    """List of urls and kernels for the Cassini mission

    More data on Cassini is available:
    https://naif.jpl.nasa.gov/pub/naif/CASSINI/
    """

    def __init__(self):
        """Initialize the Cassini kernel class
        """
        self.Lsk_url = 'https://naif.jpl.nasa.gov/pub/naif/generic_kernels/lsk/a_old_versions/naif0011.tls'
        self.Sclk_url = 'https://naif.jpl.nasa.gov/pub/naif/CASSINI/kernels/sclk/cas00171.tsc'
        self.Pck_url = 'https://naif.jpl.nasa.gov/pub/naif/CASSINI/kernels/pck/cpck09May2017.tpc'
        self.Fk_url = 'https://naif.jpl.nasa.gov/pub/naif/CASSINI/kernels/fk/release.11/cas_v37.tf'
        self.Ck_url = 'https://naif.jpl.nasa.gov/pub/naif/CASSINI/kernels/ck/04135_04171pc_psiv2.bc'
        self.Spk_url = 'https://naif.jpl.nasa.gov/pub/naif/CASSINI/kernels/spk/981005_PLTEPH-DE405S.bsp'
        self.Ik_url = 'https://naif.jpl.nasa.gov/pub/naif/CASSINI/kernels/ik/release.11/cas_iss_v09.ti'
        self.TourSpk_url = 'https://naif.jpl.nasa.gov/pub/naif/CASSINI/kernels/spk/030201AP_SK_SM546_T45.bsp'
        self.satSpk_url = 'https://naif.jpl.nasa.gov/pub/naif/CASSINI/kernels/spk/020514_SE_SAT105.bsp'
        
        self.Lsk = getPathfromUrl(self.Lsk_url)
        self.Sclk = getPathfromUrl(self.Sclk_url)
        self.Pck = getPathfromUrl(self.Pck_url)
        self.Fk = getPathfromUrl(self.Fk_url)
        self.Ck = getPathfromUrl(self.Ck_url)
        self.Spk = getPathfromUrl(self.Spk_url)
        self.Ik = getPathfromUrl(self.Ik_url)
        self.TourSpk = getPathfromUrl(self.TourSpk_url)
        self.satSpk = getPathfromUrl(self.satSpk_url)
        
        self.urlList = [self.Lsk_url, self.Sclk_url, self.Pck_url, self.Fk_url, 
                self.Ck_url, self.Spk_url, self.Ik_url, self.TourSpk_url,
                self.satSpk_url]
        self.kernelList = [self.Lsk, self.Sclk, self.Pck, self.Fk, self.Ck,
                    self.Spk, self.Ik, self.TourSpk, self.satSpk]
        self.nameList = [getKernelNameFromUrl(url) for url in self.urlList]

        self.kernelDescription = 'Metal Kernel for Cassini Orbiter'
        getKernels(self)
        self.metakernel = writeMetaKernel(self, 'cassini.tm')

def cleanupKernels(kernelObj=CassiniKernels):
    """Delete all the Kernels
    """
    for kernel in kernelObj.kernelList:
        if os.path.exists(kernel):
            os.remove(kernel)
        else:
            print("Path doesn't exist: {}".format(kernel))


def attemptDownload(url, kernelName, targetFileName, num_attempts=5):
    """Download the file from a specific url
    """
    current_attempt = 0
    while current_attempt < num_attempts:
        try:
            print("Attempting to download kernel: {}".format(kernelName))
            urlretrieve(url, targetFileName)
            break
        except:
            pass
        current_attempt = current_attempt + 1
        print("Attempting to download kernel again...")
        time.sleep(2 + current_attempt)

    if current_attempt >= num_attempts:
        print("Error downloading kernel: {}. Check if it exists at url: {}".format(kernelName, url))


def getKernels(kernelObj=CassiniKernels):
    """Download all the Kernels
    """
        
    for url in kernelObj.urlList:
        kernelName = getKernelNameFromUrl(url)
        kernelFile = os.path.join(cwd, directory, kernelName)

        if not os.path.isfile(kernelFile):
            attemptDownload(url, kernelName, kernelFile, 5)

    return 0

def writeMetaKernel(kernelObj, filename='testKernel.tm'):
    """Write a user defined meta kernel file
    """
    with open(os.path.join(cwd, directory, filename), 'w') as metaKernel:
        metaKernel.write('\\begintext\n\n')
        metaKernel.write('Created: Shankar Kulumani\n')
        if kernelObj.kernelDescription:
            metaKernel.write('Description: {}\n'.format(kernelObj.kernelDescription))

        metaKernel.write('\n')

        metaKernel.write('\\begindata\n\n')
        metaKernel.write('PATH_VALUES = (\n')
        metaKernel.write('\'{0}\'\n'.format(os.path.join(cwd, directory)))
        metaKernel.write(')\n\n')

        metaKernel.write('PATH_SYMBOLS = (\n')
        metaKernel.write('\'KERNELS\'\n')
        metaKernel.write(')\n\n')

        metaKernel.write('KERNELS_TO_LOAD = (\n')
        for kernel in kernelObj.nameList:
            metaKernel.write('\'$KERNELS/{0}\'\n'.format(kernel))

        metaKernel.write(')\n')
        metaKernel.close()

    return os.path.join(cwd, directory, filename)

if __name__ == '__main__':
    pass 



"""Time

This module contains several useful time conversion functions.

Notes
-----
You can use this module by importing it into your code

from astro import time

Attributes
----------
No module level attributes

Author
------
Shankar Kulumani		GWU		skulumani@gwu.edu
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
from kinematics import attitude


def date2jd(yr, mon, day, hr, minute, sec):
    """Convert date to Julian Date

    Purpose:
        - Converts UTC date to julian date valid 1 Mar 1900 to 28 Feb 2100

    JD = date2jd(yr, mon, day, hr, min, sec)

    Inputs:
        - yr - 4 digit year (between 1900 and 2100)
        - mon - month (between 01 and 12)
        - day - day (between 1 and 31)
        - hr - UT hour (between 0 and 23)
        - min - UT min (between 0 and 59)
        - sec - UT sec (between 0 and 59.999)

    Outputs:
        - JD - julian date (days from 1 Jan 4713 BC 12 Noon)
        - MJD - modified julian date

    Dependencies:
        - none

    Author:
        - Shankar Kulumani 21 Oct 2012
            - list revisions
        - Shankar Kulumani 3 Dec 2016
            - convert to python

    References
        - Vallado
        - USAFA Astro 321
    """

    JD = 367.0 * yr - np.floor((7 * (yr + np.floor((mon + 9) / 12.0))) *
                               0.25) + np.floor(275 * mon / 9.0) + day + 1721013.5 + (
        (sec / 60.0 + minute) / 60.0 + hr) / 24.0

    MJD = JD - 2400000.5

    return (JD, MJD)


def dayofyr2mdhms(yr, days):
    """This function converts the day of the year, days, to the month
        day, hour, minute and second.

    Algorithm     : Set up array for the number of days per month
                    loop through a temp value while the value is < the days
                    Perform integer conversions to the correct day and month
                    Convert remainder into H M S using type conversions

    Author        : Capt Dave Vallado  USAFA/DFAS     719-472-4109  26 Feb 1990
    In Ada        : Dr Ron Lisowski    USAFA/DFAS     719-472-4110  17 May 1995
    In MATLAB     : LtCol Thomas L. Yoder USAFA/DFAS  719-472-4110  Spring 00
    In Python     : Shankar Kulumani   GWU            630-336-6257  2017 06 15

    Inputs        :
        Yr          - Year                                 1900 .. 2100
        Days        - Julian Day of the year               1.0  .. 366.0

    OutPuts       :
        Mon         - Month                                   1 .. 12
        D           - Day                                     1 .. 28,29,30,31
        H           - Hour                                    0 .. 23
        M           - Minute                                  0 .. 59
        Sec         - Second                                0.0 .. 59.999

    Locals        :
        DayOfyr     - Day of year
        Temp        - Temporary Long_Float value
        IntTemp     - Temporary 16 bit Integer value
        i           - Index

    Constants     :
        LMonth(12)  - Integer Array containing the number of days per month

    Coupling      : None

    References    :
        None.
    """
    dayofyr = np.fix(days)

    # find month and day of month
    lmonth = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    if (yr - 1900) % 4 == 0:
        lmonth[1] = 29

    ii = 0
    temp = 0
    while (dayofyr > temp + lmonth[ii]) and (ii < 11):
        temp += lmonth[ii]
        ii += 1

    mon = ii + 1
    day = dayofyr - temp

    # find hour, min, and seconds
    temp = (days - dayofyr) * 24
    hour = np.fix(temp) - 0
    temp = (temp - hour) * 60.0
    minute = np.fix(temp) - 0
    sec = (temp - minute) * 60.0

    return (mon, day, hour, minute, sec)


def jd2date(jd):
    """This function finds the Year, month, day, hour, minute and second
    given the Julian date.

    Algorithm     : Set up starting values
                    Find the elapsed days through the year in a loop
                    Call routine to find each individual value

    Author        : Capt Dave Vallado  USAFA/DFAS  719-472-4109  26 Feb 1990
    In Ada        : Dr Ron Lisowski    USAFA/DFAS  719-472-4110  17 May 1995
    In Matlab     : LtCol Thomas Yoder USAFA/DFAS  719-333-4110  Spring 2000
    In Python     : Shankar Kulumani   GWU         630-336-6257  2017 06 15

    Inputs        :
        JD          - Julian Date                          days from 4713 B.C.

    OutPuts       :
        Yr          - Year                                 1900 .. 2100
        Mon         - Month                                   1 .. 12
        D           - Day                                     1 .. 28,29,30,31
        H           - Hour                                    0 .. 23
        M           - Minute                                  0 .. 59
        S           - Second                                0.0 .. 59.999

    Locals        :
        days        - Day of year plus fraction of a day   days
        Tu          - Julian Centuries from 1 Jan 1900
        Temp        - Temporary Long_Float value
        LeapYrs     - Number of Leap years from 1900

    Constants     :
        None.

    Coupling      :
        DayofYr2MDHMS  - Finds Month, day, hour, minute and second given Days
        and Yr

    References    :
        1988 Almanac for Computers  pg. B2
        Escobal       pg. 17-19
        Kaplan        pg. 329-330
    """

    temp = jd - 2415019.5
    tu = temp / 365.25
    yr = 1900 + np.fix(tu)
    leapyrs = np.fix((yr - 1900 - 1) * 0.25)
    days = temp - ((yr - 1900) * 365.0 + leapyrs)

    # check for beginning of year
    if days < 1.0:
        yr = yr - 1
        leapyrs = np.fix((yr - 1900 - 1) * 0.25)
        days = temp - ((yr - 1900) * 365.0 + leapyrs)

    mon, day, h, m, s = dayofyr2mdhms(yr, days)

    return (yr, mon, day, h, m, s)


def gsttime0(yr):
    """This function finds the Greenwich Sidereal time at the beginning of a
    year.  This formula is derived from the Astronomical Almanac and is good
    only for 0 hr UT, 1 Jan of a year.

    Algorithm     : Find the Julian Date Ref 4713 BC
                    Perform expansion calculation to obtain the answer
                    Check the answer for the correct quadrant and size

    Author        : Capt Dave Vallado  USAFA/DFAS  719-472-4109   12 Feb 1989
    In Ada        : Dr Ron Lisowski    USAFA/DFAS  719-472-4110   17 May 1995
    In MatLab     : LtCol Thomas Yoder USAFA/DFAS  719-333-4110   Spring 2000
    Remove Global : Shankar Kulumani 7 Dec 2014
    In Python     : Shankar Kulumani   GWU         630-336-6257   10 Jun 2017

    Inputs        :
        Yr          - Year                                 1988, 1989, etc.

    OutPuts       :
        GST0        - Returned Greenwich Sidereal Time     0 to 2Pi rad

    Locals        :
        JD          - Julian Date                          days from 4713 B.C.
        Tu          - Julian Centuries from 1 Jan 2000

    Constants     :
        TwoPI         Two times Pi (DFASmath.m constant)

    Coupling      :
        mod			  MatLab modulus function

    References    :
        1989 Astronomical Almanac pg. B6
        Escobal       pg. 18 - 21
        Explanatory Supplement pg. 73-75
        Kaplan        pg. 330-332
        BMW           pg. 103-104
    """

    TwoPI = 2 * np.pi

    JD = (367.0 * yr - np.fix(7.0 * (yr + np.fix(10.0 / 12.0)) * 0.25) +
          np.fix(275.0 / 9.0) + 1721014.5)
    Tu = (np.fix(JD) + 0.5 - 2451545.0) / 36525.0
    GST0 = 1.753368559 + 628.3319705 * Tu + 6.770708127E-06 * Tu * Tu
    GST0 = attitude.normalize(GST0, 0, 2 * np.pi)
    return GST0


def gstlst(jd, site_lon):
    """This program calculates GST and LST given the Julian Day and site
    longitude.

    Author:  Shankar Kulumani GWU 18 Jun 2017

    Inputs:
        jd : float
            Julian Day
        sitlon : site longitude (radians)
    Outputs:
        gst - Greenwich Sidereal Time
        lst - Local Sidereal Time

    Constants: None

    Coupling: 

    Modifications:
        18 Jun 2017 - use algorithm 15 from Vallado

    References:
        Astro 321 PREDICT
        Vallado Algorithm 15 
    """
    deg2rad = np.pi / 180
    hour2sec = 3600
    sec2deg = 15 / 3600

    Tut1 = (jd - 2451545.0) / 36525

    gst = (- 6.2e-6 * Tut1 * Tut1 * Tut1 + 0.093104 * Tut1 * Tut1
           + (876600.0 * 3600.0 + 8640184.812866) * Tut1 + 67310.54841)
    gst = (gst % 86400) * sec2deg

    gst = attitude.normalize(gst * deg2rad, 0, 2 * np.pi)
    lst = attitude.normalize(gst[0] + site_lon, 0, 2 * np.pi)

    return gst[0], lst[0]


def finddays(yr, mo, day, hr, m, sec):
    """This function finds the fractional days through a year given the year,
        month, day, hour, minute and second. Midnight New Year's is 0.0

    Algorithm     : Set up array for the number of days per month
                    Check for a leap year
                    Loop to find the elapsed days in the year

    Author        : Capt Dave Vallado  USAFA/DFAS  719-472-4109  11 Dec 1990
    In Ada        : Dr Ron Lisowski    USAFA/DFAS  719-472-4110  17 May 1995
    In MatLab     : LtCol Thomas Yoder USAFA/DFAS  719-333-4110  Spring 2000

    Inputs        :
        Year        - Year                                 1900 .. 2100
        Month       - Month                                   1 .. 12
        Day         - Day                                     1 .. 28,29,30,31
        Hr          - Hour                                    0 .. 23
        Min         - Minute                                  0 .. 59
        Sec         - Second                                0.0 .. 59.999

    OutPuts       :
        DDays        - Fractional elapsed day of year    days

    Locals        :
        i           - Index
            LMonth	  - array holding number of days in each month

    Constants     :
        None.

    Coupling      :
        None.

    """

    LMonth = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    if (yr - 1900) % 4 == 0:
        LMonth[1] = 29

    i = 1
    DDays = 0.0
    while (i < mo) and (i < 11):
        DDays = DDays + LMonth[i]
        i = i + 1

    DDays = DDays + (day - 1) + hr / 24.0 + m / 1440.0 + sec / 86400.0

    return DDays


if __name__ == "__main__":
    # test JD for J2000
    yr = 2000
    mon = 1
    day = 1
    hour = 12
    minute = 0
    sec = 0

    JD, MJD = date2jd(yr, mon, day, hour, minute, sec)

    print("J2000: %9.2f" % JD)



"""Coordinate transformations

"""

import numpy as np
from astro import kepler, constants
from kinematics import attitude

def dcm_pqw2eci_vector(r, v, mu=constants.earth.mu):
    """Define rotation matrix transforming PQW to ECI given eci vectors

    """
    h_vec, _, ecc_vec = kepler.hne_vec(r, v, mu)
    p_hat = ecc_vec / np.linalg.norm(ecc_vec)
    w_hat = h_vec / np.linalg.norm(h_vec) 
    q_hat = np.cross(w_hat, p_hat)

    dcm = np.stack((p_hat, q_hat, w_hat), axis=1)

    return dcm

def dcm_pqw2eci_coe(raan, inc, arg_p):
    """Define rotation matrix transforming PQW to ECI given orbital elements

    """
    dcm = attitude.rot3(raan).dot(
            attitude.rot1(inc)).dot(attitude.rot3(arg_p))

    return dcm
